# -*- coding: utf-8 -*-
from operator import attrgetter
import datetime

import numpy as np

from ..distance import DistanceHypothesiser
from ...types.detection import Detection
from ...types.state import GaussianState
from ...types.track import Track
from ... import measures


def test_mahalanobis(predictor, updater):

    timestamp = datetime.datetime.now()
    track = Track([GaussianState(np.array([[0]]), np.array([[1]]), timestamp)])
    detection1 = Detection(np.array([[2]]))
    detection2 = Detection(np.array([[3]]))
    detections = {detection1, detection2}

    measure = measures.Mahalanobis()
    hypothesiser = DistanceHypothesiser(predictor, updater, measure=measure)

    hypotheses = hypothesiser.hypothesise(track, detections, timestamp)

    # There are 3 hypotheses - Detection 1, Detection 2, Missed Dectection
    assert len(hypotheses) == 3

    # There is a missed detection hypothesis
    assert any(not hypothesis.measurement for hypothesis in hypotheses)

    # Each hypothesis has a distance attribute
    assert all(hypothesis.distance >= 0 for hypothesis in hypotheses)

    # The hypotheses are sorted correctly
    assert min(hypotheses, key=attrgetter('distance')) is hypotheses[0]



# -*- coding: utf-8 -*-
from abc import abstractmethod

import scipy as sp

from ..base import Model
from ...base import Property


class ControlModel(Model):
    """Control Model base class"""

    ndim_state = Property(int, doc="Number of state dimensions")
    mapping = Property(
        sp.ndarray, doc="Mapping between control and state dims")

    @property
    @abstractmethod
    def ndim_ctrl(self):
        """Number of control input dimensions"""
        pass



# -*- coding: utf-8 -*-
import datetime

import numpy as np
from scipy.spatial import distance

from .. import measures

from ..types.array import StateVector, CovarianceMatrix
from ..types.state import GaussianState

# Create a time stamp to use for both states
t = datetime.datetime.now()

# Set target ground truth prior
u = StateVector([[10], [1], [10], [1]])
ui = CovarianceMatrix(np.diag([100, 10, 100, 10]))

state_u = GaussianState(u, ui, timestamp=t)

v = StateVector([[11], [10], [100], [2]])
vi = CovarianceMatrix(np.diag([20, 3, 7, 10]))

state_v = GaussianState(v, vi, timestamp=t)


def test_euclidean():

    measure = measures.Euclidean()
    assert measure(state_u, state_v) == distance.euclidean(u, v)


def test_euclideanweighted():
    weight = np.array([[1], [2], [3], [1]])
    measure = measures.EuclideanWeighted(weight)
    assert measure(state_u, state_v) == distance.euclidean(u, v, weight)


def test_mahalanobis():
    measure = measures.Mahalanobis()
    assert measure(state_u, state_v) == distance.mahalanobis(u,
                                                             v,
                                                             np.linalg.inv(ui))


def test_hellinger():
    measure = measures.GaussianHellinger()
    # Distance value obtained from MATLAB
    assert np.isclose(measure(state_u, state_v), 0.665, atol=1e-3)


def test_zero_hellinger():
    measure = measures.GaussianHellinger()
    # Set target ground truth prior
    u = StateVector([[10], [1], [10], [1]])
    ui = CovarianceMatrix(np.diag([0, 0, 0, 0]))
    state_u = GaussianState(u, ui, timestamp=t)

    v = StateVector([[11], [10], [100], [2]])
    vi = CovarianceMatrix(np.diag([0, 0, 0, 0]))
    state_v = GaussianState(v, vi, timestamp=t)
    print(measure(state_u, state_v))
    # Distance value obtained from MATLAB
    assert np.isclose(measure(state_u, state_v), 1, atol=1e-3)


def test_squared_hellinger():
    measure = measures.SquaredGaussianHellinger()
    # Distance value obtained from MATLAB
    assert np.isclose(measure(state_u, state_v), 0.444, atol=1e-3)


def test_hellinger_full_mapping():
    mapping = np.arange(len(u))
    measure = measures.GaussianHellinger(mapping=mapping)
    assert np.isclose(measure(state_u, state_v), 0.665, atol=1e-3)


def test_hellinger_partial_mapping():
    # Distance value obtained from MATLAB
    mapping = np.array([0, 1])
    measure = measures.GaussianHellinger(mapping=mapping)
    assert np.isclose(measure(state_u, state_v), 0.4555, atol=1e-3)
    mapping = np.array([0, 3])
    measure = measures.GaussianHellinger(mapping=mapping)
    assert np.isclose(measure(state_u, state_v), 0.3701, atol=1e-3)


def test_mahalanobis_full_mapping():
    mapping = np.arange(len(u))
    measure = measures.Mahalanobis(mapping=mapping)
    assert measure(state_u, state_v) == distance.mahalanobis(u,
                                                             v,
                                                             np.linalg.inv(ui))


def test_mahalanobis_partial_mapping():
    mapping = np.array([0, 1])
    measure = measures.Mahalanobis(mapping=mapping)
    reduced_ui = CovarianceMatrix(np.diag([100, 10]))
    assert measure(state_u, state_v) == \
        distance.mahalanobis([[10], [1]],
                             [[11], [10]], np.linalg.inv(reduced_ui))
    mapping = np.array([0, 3])
    reduced_ui = CovarianceMatrix(np.diag([100, 10]))
    measure = measures.Mahalanobis(mapping=mapping)
    assert measure(state_u, state_v) == \
        distance.mahalanobis([[10], [1]],
                             [[11], [2]], np.linalg.inv(reduced_ui))


def test_euclidean_full_mapping():
    mapping = np.arange(len(u))
    measure = measures.Euclidean(mapping=mapping)
    assert measure(state_u, state_v) == distance.euclidean(u, v)


def test_euclidean_partial_mapping():
    mapping = np.array([0, 1])
    measure = measures.Euclidean(mapping=mapping)
    assert measure(state_u, state_v) == \
        distance.euclidean([[10], [1]], [[11], [10]])
    mapping = np.array([0, 3])
    measure = measures.Euclidean(mapping=mapping)
    assert measure(state_u, state_v) == \
        distance.euclidean([[10], [1]], [[11], [2]])


def test_euclideanweighted_full_mapping():
    weight = np.array([[1], [2], [3], [1]])
    mapping = np.arange(len(u))
    measure = measures.EuclideanWeighted(weight, mapping=mapping)
    assert measure(state_u, state_v) == distance.euclidean(u, v, weight)


def test_euclideanweighted_partial_mapping():
    weight = np.array([[1], [2]])
    mapping = np.array([0, 1])
    measure = measures.EuclideanWeighted(weight, mapping=mapping)
    assert measure(state_u, state_v) == \
        distance.euclidean([[10], [1]], [[11], [10]], weight)
    mapping = np.array([0, 3])
    measure = measures.EuclideanWeighted(weight, mapping=mapping)
    assert measure(state_u, state_v) == \
        distance.euclidean([[10], [1]], [[11], [2]], weight)



# -*- coding: utf-8 -*-
import bisect
import datetime
from warnings import warn

from ..base import Property
from .base import Feeder


class TimeBufferedFeeder(Feeder):
    """Buffer detections so they can be yielded in time order.

    Any "old" detections (where the time is earlier than the head of the
    buffer) shall be dropped, producing a :class:`UserWarning`.
    """
    buffer_size = Property(int, default=1000, doc="Max size of buffer")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._detections = set()

    @property
    def detections(self):
        return self._detections

    def detections_gen(self):
        detections_iter = iter(self.detector.detections_gen())
        time_detections_buffer = [next(detections_iter)]

        for time_detections in detections_iter:
            # Drop "old" detections
            if len(time_detections_buffer) >= self.buffer_size and \
                    time_detections < time_detections_buffer[0]:
                warn('"Old" detection dropped')
                continue

            # Insert in order
            bisect.insort(time_detections_buffer, time_detections)

            # Yield oldest when buffer full
            if len(time_detections_buffer) > self.buffer_size:
                time, detections = time_detections_buffer.pop(0)
                self._detections = detections
                yield time, detections

        # No more new detections: yield remaining buffer
        for time, detections in time_detections_buffer:
            self._detections = detections
            yield time, detections


class TimeSyncFeeder(Feeder):
    """Synchronise the detections into selected time window.

    Assumes that detections from :attr:`detector` are in time order. The
    :class:`~.TimeBufferedFeeder` can be used in conjunction to ensure this is
    the case.
    """

    time_window = Property(
        datetime.timedelta,
        default=datetime.timedelta(seconds=1),
        doc="Time window to group detections")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._detections = set()

    @property
    def detections(self):
        return self._detections

    def detections_gen(self):
        detections_iter = iter(self.detector.detections_gen())
        prev_time, detections = next(detections_iter)
        detections_buffer = set(detections)

        for time, detections in detections_iter:
            if time > prev_time + self.time_window:
                self._detections = detections_buffer
                yield prev_time + self.time_window, self.detections

                # Increment time and yield empty detections set until latest
                # detections within time window.
                prev_time += self.time_window
                while time > prev_time + self.time_window:
                    self._detections = set()
                    prev_time += self.time_window
                    yield prev_time, self.detections

                detections_buffer = set(detections)
            else:
                detections_buffer.update(detections)

        # No more new detections: yield remaining buffer
        self._detections = detections_buffer
        yield prev_time + self.time_window, self.detections



# coding: utf-8
import datetime
import pytest
import numpy as np

from ...models.transition.linear import ConstantVelocity
from ...predictor.kalman import (
    KalmanPredictor, ExtendedKalmanPredictor, UnscentedKalmanPredictor)
from ...types.prediction import GaussianStatePrediction
from ...types.state import GaussianState


@pytest.mark.parametrize(
    "PredictorClass, transition_model, prior_mean, prior_covar",
    [
        (   # Standard Kalman
            KalmanPredictor,
            ConstantVelocity(noise_diff_coeff=0.1),
            np.array([[-6.45], [0.7]]),
            np.array([[4.1123, 0.0013],
                      [0.0013, 0.0365]])
        ),
        (   # Extended Kalman
            ExtendedKalmanPredictor,
            ConstantVelocity(noise_diff_coeff=0.1),
            np.array([[-6.45], [0.7]]),
            np.array([[4.1123, 0.0013],
                      [0.0013, 0.0365]])
        ),
        (   # Unscented Kalman
            UnscentedKalmanPredictor,
            ConstantVelocity(noise_diff_coeff=0.1),
            np.array([[-6.45], [0.7]]),
            np.array([[4.1123, 0.0013],
                      [0.0013, 0.0365]])
        )
    ],
    ids=["standard", "extended", "unscented"]
)
def test_kalman(PredictorClass, transition_model,
                prior_mean, prior_covar):

    # Define time related variables
    timestamp = datetime.datetime.now()
    timediff = 2  # 2sec
    new_timestamp = timestamp + datetime.timedelta(seconds=timediff)
    time_interval = new_timestamp - timestamp

    # Define prior state
    prior = GaussianState(prior_mean,
                          prior_covar,
                          timestamp=timestamp)

    # Calculate evaluation variables
    eval_prediction = GaussianStatePrediction(
        transition_model.matrix(timestamp=new_timestamp,
                                time_interval=time_interval)@prior.mean,
        transition_model.matrix(timestamp=new_timestamp,
                                time_interval=time_interval)
        @prior.covar
        @transition_model.matrix(timestamp=new_timestamp,
                                 time_interval=time_interval).T
        + transition_model.covar(timestamp=new_timestamp,
                                 time_interval=time_interval))

    # Initialise a kalman predictor
    predictor = PredictorClass(transition_model=transition_model)

    # Perform and assert state prediction
    prediction = predictor.predict(prior=prior,
                                   timestamp=new_timestamp)

    assert(np.allclose(prediction.mean,
                       eval_prediction.mean, 0, atol=1.e-14))
    assert(np.allclose(prediction.covar,
                       eval_prediction.covar, 0, atol=1.e-14))
    assert(prediction.timestamp == new_timestamp)

    # TODO: Test with Control Model



# -*- coding: utf-8 -*-
from .base import MetricManager
from ..base import Property
from ..dataassociator import Associator
from ..types.detection import Detection
from ..types.groundtruth import GroundTruthPath
from ..types.track import Track


class SimpleManager(MetricManager):
    """SimpleManager class for metric management

    Simple :class:`~.MetricManager` for the generation of metrics on multiple
    :class:`~.Track`, :class:`~.Detection` and :class:`~.GroundTruthPath`
    objects.
    """
    generators = Property(list, doc='List of generators to use', default=None)
    associator = Property(Associator,
                          doc="Associator to combine tracks and truth",
                          default=None)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tracks = set()
        self.groundtruth_paths = set()
        self.detections = set()
        self.association_set = None

    def add_data(self, input_objects, overwrite=True):
        """Adds data to the metric generator

        Parameters
        ----------
        input_objects : list or set of objects
            Objects to be added to the manager. The class of the object is used
            to determine whether it is track, truth or detection
        overwrite: bool
            declaring whether pre-existing data will be overwritten. Note that
            overwriting one field (e.g. tracks) does not affect the others
        """
        for in_obj in input_objects:
            if not isinstance(in_obj, (list, set)):
                raise TypeError('Inputs are expected as lists or sets only')
            else:
                if all(isinstance(x, Track) for x in in_obj):
                    if overwrite:
                        self.tracks = set(in_obj)
                    else:
                        self.tracks = self.tracks.union(set(in_obj))
                elif all(isinstance(x, GroundTruthPath)
                         for x in in_obj):
                    if overwrite:
                        self.groundtruth_paths = set(in_obj)
                    else:
                        self.groundtruth_paths = self.groundtruth_paths.union(
                            set(in_obj))
                elif all(isinstance(x, Detection) for x in in_obj):
                    if overwrite:
                        self.detections = set(in_obj)
                    else:
                        self.detections = self.detections.union(set(in_obj))
                else:
                    raise TypeError(
                        'Object of type {!r} not expected'.format(type()))
                    # This error doesn't work if the first element of the list
                    # is a sensible one but the later ones aren't.

    def associate_tracks(self):
        """Associate tracks to truth using the associator

        The resultant :class:`~.AssociationSet` internally.
        """

        self.association_set = self.associator.associate_tracks(
            self.tracks, self.groundtruth_paths)

    def generate_metrics(self):
        """Generate metrics using the generators and data that has been added

        Returns
        ----------
        : set of :class:`~.Metric`
            Metrics generated
        """
        metrics = set()
        for generator in self.generators:
            metric = generator.compute_metric(self)
            # Metrics can be lists or not, there's probably a neater way to do
            # this
            if isinstance(metric, list):
                metrics.update(metric)
            else:
                metrics.add(metric)

        return set(metrics)

    def list_timestamps(self):
        """List all the timestamps used in the tracks and truth, in order

        Returns
        ----------
        : list of :class:`datetime.datetime`
            unique timestamps present in the internal tracks and truths.
        """

        # Make a list of all the unique timestamps used
        timestamps = {state.timestamp for state in self.tracks}
        timestamps |= {state.timestamp
                       for path in self.groundtruth_paths
                       for state in path}
        return sorted(timestamps)



# -*- coding: utf-8 -*-

from pytest import approx
from numpy import deg2rad
import numpy as np
from ...functions import mod_elevation, mod_bearing
from math import trunc, ceil, floor

from ..angle import Bearing, Elevation


def pytest_generate_tests(metafunc):
    # Called on each function in a class
    idlist = []
    argvalues = []
    for scenario in metafunc.cls.scenarios:
        idlist.append(scenario[0])
        items = scenario[1].items()
        argnames = [x[0] for x in items]
        argvalues.append(([x[1] for x in items]))
    metafunc.parametrize(argnames, argvalues, ids=idlist, scope="class")


scenario1 = ('bearing', {'Class': Bearing, 'func': mod_bearing})
scenario2 = ('elevation', {'Class': Elevation, 'func': mod_elevation})


class TestAngle:
    scenarios = [scenario1, scenario2]
    # In the following functions:
    # Class - this is the class type we use for casting e.g. Bearing
    # func - this is the modulo function

    def test_bearing_init(self, Class, func):
        b1 = Class(3.14)
        b2 = Class(deg2rad(700))
        assert approx(b1) == func(3.14)
        assert approx(b2) == func(deg2rad(-20))
        assert b1 == b1
        assert b1 == b1
        assert b1 != b2

    def test_multiply(self, Class, func):
        b1 = Class(3.14)
        b2 = Class(2)
        bmul = float(b1)*float(b2)
        bmul2 = 2*float(b1)
        assert b1*b2 == approx(bmul)
        assert b2*b1 == approx(bmul)
        assert 2*b1 == approx(bmul2)
        assert b1*2 == approx(bmul2)

    def test_addition(self, Class, func):
        b1 = Class(3.14)
        b2 = Class(2)
        b_sum = float(b1) + float(b2)
        assert approx(func(b_sum)) == b1+b2
        assert approx(func(b_sum)) == float(b1)+b2
        assert approx(func(b_sum)) == b1+float(b2)
        assert approx(func(b_sum)) == b2+b1

    def test_subtraction(self, Class, func):
        b1 = Class(3.14)
        b2 = Class(2)
        b_diff = float(b1) - float(b2)
        assert approx(func(b_diff)) == b1-b2
        assert approx(func(b_diff)) == float(b1)-b2
        assert approx(func(b_diff)) == b1-float(b2)
        assert approx(-func(b_diff)) == b2-b1

    def test_division(self, Class, func):
        b1 = Class(3.14)
        b2 = Class(2)
        b_div = float(b1)/float(b2)
        assert approx(b_div) == b1/b2
        assert approx(float(b1)/2) == b1/2
        assert approx(b_div) == float(b1)/b2

    def test_comaparision(self, Class, func):
        b1 = Class(deg2rad(30))
        b2 = Class(deg2rad(10))
        assert b1 == b1
        assert b1 > b2
        assert b1 >= b1
        assert b1 <= b1
        assert b2 < b1

    def test_trig(self, Class, func):
        b1 = Class(deg2rad(22.0))
        b2 = float(b1)

        assert np.sin(b1) == np.sin(b2)
        assert np.cos(b1) == np.cos(b2)
        assert np.tan(b1) == np.tan(b2)
        assert np.sinh(b1) == np.sinh(b2)
        assert np.cosh(b1) == np.cosh(b2)
        assert np.tanh(b1) == np.tanh(b2)
        assert np.rad2deg(b1) == np.rad2deg(b2)

    def test_misc(self, Class, func):
        b1 = Class(deg2rad(22.0))
        b2 = float(b1)

        assert ceil(b1) == ceil(b2)
        assert floor(b1) == floor(b2)
        assert trunc(b1) == trunc(b2)
        assert round(b1, 2) == round(b2, 2)
        assert b1 % 0.1 == b2 % 0.1
        assert +b1 == +b2
        assert approx(-b1) == -b2



# -*- coding: utf-8 -*-
import warnings
from abc import abstractmethod

import utm
from pymap3d import geodetic2enu, geodetic2ned

from .base import Feeder
from ..base import Property
from ..types.detection import Detection


class _LLARefConverter(Feeder):
    reference_point = Property(
        (float, float, float), doc="(Long, Lat, Altitude)")
    mapping = Property(
        (float, float, float), default=(0, 1, 2),
        doc="Indexes of long, lat, altitude. Default (0, 1, 2)")

    @property
    @abstractmethod
    def _converter(self):
        raise NotImplementedError

    @property
    def detections(self):
        return self._detections.copy()

    def detections_gen(self):
        for time, detections in self.detector.detections_gen():

            self._detections = set()
            for detection in detections:
                new_coord = self._converter(
                    detection.state_vector[self.mapping[1], 0],  # Lat
                    detection.state_vector[self.mapping[0], 0],  # Long
                    detection.state_vector[self.mapping[2], 0],  # Altitude
                    self.reference_point[1],  # Lat
                    self.reference_point[0],  # Long
                    self.reference_point[2],  # Altitude
                )
                state_vector = detection.state_vector.copy()
                state_vector[self.mapping, 0] = new_coord
                self._detections.add(
                    Detection(
                        state_vector,
                        timestamp=detection.timestamp,
                        measurement_model=detection.measurement_model,
                        metadata=detection.metadata))

            yield time, self.detections


class LLAtoENUConverter(_LLARefConverter):
    """Converts Long., Lat., Alt. to East, North, Up coordinate space.

    This replaces Longitude (°), Latitude (°) and Altitude (m) of a
    :class:`~.Detection` with East (m), North (m) and Up (m) coordinate space
    from a defined :attr:`reference_point`.
    """

    @property
    def _converter(self):
        return geodetic2enu


class LLAtoNEDConverter(_LLARefConverter):
    """Converts Long., Lat., Alt. to North, East, Down coordinate space.

    This replaces Longitude (°), Latitude (°) and Altitude (m) of a
    :class:`~.Detection` with North (m), East (m) and Down (m) coordinate space
    from a defined :attr:`reference_point`.
    """

    @property
    def _converter(self):
        return geodetic2ned


class LongLatToUTMConverter(Feeder):
    """Converts long. and lat. to UTM easting and northing.

    This replaces Longitude (°), Latitude (°) of a :class:`~.Detection` with
    East (m), North (m) coordinate space in a Universal Transverse Mercator
    zone.

    Note
    ----
    Whilst this allows for detections to be converted even if they are
    outside of the set zone, the errors will increase the further the
    position is from the zone being used.

    """

    mapping = Property(
        (float, float), default=(0, 1),
        doc="Indexes of long, lat. Default (0, 1)")
    zone_number = Property(
        int, default=None,
        doc="UTM zone number to carry out conversion. Default `None`, where it"
            "will select the zone based on the first detection.")
    northern = Property(
        bool, default=None,
        doc="UTM northern for northern or southern grid. Default `None`, where"
            "it will be base on the first detection.")

    @property
    def detections(self):
        return self._detections.copy()

    def detections_gen(self):
        for time, detections in self.detector.detections_gen():

            self._detections = set()
            for detection in detections:
                easting, northing, zone_num, northern = utm.from_latlon(
                    *detection.state_vector[self.mapping[::-1], 0],
                    self.zone_number)
                if self.zone_number is None:
                    self.zone_number = zone_num
                if self.northern is None:
                    self.northern = northern >= 'N'
                elif (self.northern and northern < 'N') or (
                            not self.northern and northern >= 'N'):
                    warnings.warn("Detection cannot be converted to UTM zone")
                    continue

                state_vector = detection.state_vector.copy()
                state_vector[self.mapping, 0] = easting, northing
                self._detections.add(
                    Detection(
                        state_vector,
                        timestamp=detection.timestamp,
                        measurement_model=detection.measurement_model,
                        metadata=detection.metadata))

            yield time, self.detections



# -*- coding: utf-8 -*-
from abc import abstractmethod

from ..base import Base, Property
from ..models.measurement import MeasurementModel


class Updater(Base):
    """Updater base class

    An updater is used to update the state, utilising a
    :class:`~.MeasurementModel`.
    """

    measurement_model = Property(MeasurementModel, doc="measurement model")

    @abstractmethod
    def get_measurement_prediction(
            self, state_prediction, measurement_model=None, **kwargs):
        """Get measurement prediction from state prediction

        Parameters
        ----------
        state_prediction : :class:`~.StatePrediction`
            The state prediction
        measurement_model: :class:`~.MeasurementModel`, optional
            The measurement model used to generate the measurement prediction.
            Should be used in cases where the measurement model is dependent
            on the received measurement. The default is `None`, in which case
            the updater will use the measurement model specified on
            initialisation


        Returns
        -------
        : :class:`~.MeasurementPrediction`
            The predicted measurement
        """
        raise NotImplementedError

    @abstractmethod
    def update(self, hypothesis, **kwargs):
        """Update state using prediction and measurement.

        Parameters
        ----------
        hypothesis : :class:`~.Hypothesis`
            Hypothesis with predicted state and associated detection used for
            updating.

        Returns
        -------
        : :class:`~.State`
            The state posterior
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
import numpy as np
from numbers import Real
from numpy import float64
from math import trunc, ceil, floor
from ..functions import mod_bearing, mod_elevation


class Angle(Real):
    """Angle class.

    Angle handles modulo arithmetic for adding and subtracting angles
    """
    @staticmethod
    def mod_angle(value):
        return value

    def __init__(self, value):
        self._value = float64(self.mod_angle(value))

    def __add__(self, other):
        out = self._value + float64(other)
        return self.__class__(self.mod_angle(out))

    def __radd__(self, other):
        return self.__class__.__add__(self, other)

    def __sub__(self, other):
        out = self._value - float64(other)
        return self.__class__(self.mod_angle(out))

    def __rsub__(self, other):
        return self.__class__.__add__(-self, other)

    def __float__(self):
        return float(self._value)

    def __mul__(self, other):
        return self._value * float64(other)

    def __rmul__(self, other):
        return self._value * float64(other)

    def __str__(self):
        return str(self._value)

    def __repr__(self):
        return "{0}({1!r})".format(self.__class__.__name__, float64(self))

    def __neg__(self):
        return self.__class__(-self._value)

    def __truediv__(self, other):
        return self._value / float64(other)

    def __rtruediv__(self, other):
        return float64(other) / self._value

    def __eq__(self, other):
        return self._value == float64(other)

    def __ne__(self, other):
        return self._value != float64(other)

    def __abs__(self):
        return self.__class__(abs(self._value))

    def __le__(self, other):
        return self._value <= float64(other)

    def __lt__(self, other):
        return self._value < float64(other)

    def __ge__(self, other):
        return self._value >= float64(other)

    def __gt__(self, other):
        return self._value > float64(other)

    def __floor__(self):
        return floor(self._value)

    def __ceil__(self):
        return ceil(self._value)

    def __floordiv__(self, other):
        return self._value // other

    def __mod__(self, other):
        return self._value % other

    def __pos__(self):
        return self.__class__(+self._value)

    def __pow__(self, value):
        return pow(self._value, value)

    def __rfloordiv__(self, other):
        return other // self._value

    def __rmod__(self, other):
        return other % self._value

    def __round__(self, ndigits=None):
        return float64(round(self._value, ndigits=ndigits))

    def __rpow__(self, base):
        return NotImplemented

    def __trunc__(self):
        return trunc(self._value)

    def cos(self):
        return np.cos(self._value)

    def sin(self):
        return np.sin(self._value)

    def tan(self):
        return np.tan(self._value)

    def cosh(self):
        return np.cosh(self._value)

    def sinh(self):
        return np.sinh(self._value)

    def tanh(self):
        return np.tanh(self._value)

    def rad2deg(self):
        return np.rad2deg(self._value)


class Bearing(Angle):
    """Bearing angle class.

    Bearing handles modulo arithmetic for adding and subtracting angles. \
    The retuern type for addition and subtraction is Bearing.
    Multiplcation or division produces a float object rather than Bearing.
    """
    @staticmethod
    def mod_angle(value):
        return mod_bearing(value)


class Elevation(Angle):
    """Elevation angle class.

    Elevation handles modulo arithmetic for adding and subtracting elevation
    angles. The retuern type for addition and subtraction is Elevation.
    Multiplcation or division produces a float object rather than Elevation.
    """
    @staticmethod
    def mod_angle(value):
        return mod_elevation(value)



# -*- coding: utf-8 -*-
"""Provides an ability to serialise Stone Soup objects into and from YAML.

Stone Soup utilises YAML_ for serialisation. The :doc:`stonesoup.base`
feature of components is exploited in order to store the data of the
components and data types.

.. _YAML: http://yaml.org/"""
import datetime
import warnings
from io import StringIO
from collections import OrderedDict
from functools import lru_cache
from pathlib import Path
from importlib import import_module

import numpy as np
import ruamel.yaml
from ruamel.yaml.constructor import ConstructorError

from .base import Base


class YAML:
    """Class for YAML serialisation."""
    tag_prefix = '!{}.'.format(__name__.split('.', 1)[0])

    def __init__(self):
        self._yaml = ruamel.yaml.YAML()
        self._yaml.default_flow_style = False

        # NumPy
        self._yaml.representer.add_multi_representer(
            np.ndarray, self.ndarray_to_yaml)
        self._yaml.constructor.add_constructor(
            "!numpy.ndarray", self.ndarray_from_yaml)

        # Datetime
        self._yaml.representer.add_representer(
            datetime.timedelta, self.timedelta_to_yaml)
        self._yaml.constructor.add_constructor(
            "!datetime.timedelta", self.timedelta_from_yaml)

        # Path
        self._yaml.representer.add_multi_representer(
            Path, self.path_to_yaml)
        self._yaml.constructor.add_constructor(
            "!pathlib.Path", self.path_from_yaml)

        # Declarative classes
        self._yaml.representer.add_multi_representer(
            Base, self.declarative_to_yaml)
        self._yaml.constructor.add_multi_constructor(
            self.tag_prefix, self.declarative_from_yaml)

    def dump(self, data, stream, **kwargs):
        return self._yaml.dump(data, stream, **kwargs)

    def dumps(self, data, *args, **kwargs):
        """Return as a string."""
        stream = StringIO()
        self.dump(data, stream, *args, **kwargs)
        return stream.getvalue()

    def dump_all(self, documents, stream, **kwargs):
        return self._yaml.dump_all(self, documents, stream, **kwargs)

    def load(self, stream):
        return self._yaml.load(stream)

    def load_all(self, stream):
        yield from self._yaml.load_all(stream)

    @classmethod
    def yaml_tag(cls, class_):
        """Return YAML tag for object.

        Constructed from module and class name."""
        return "!{}.{}".format(class_.__module__, class_.__qualname__)

    @classmethod
    def declarative_to_yaml(cls, representer, node):
        """Convert declarative class instances to YAML.

        Store as mapping of declared properties, skipping any which are the
        default value."""
        return representer.represent_omap(
            cls.yaml_tag(type(node)),
            OrderedDict((name, getattr(node, name))
                        for name, property_ in type(node).properties.items()
                        if getattr(node, name) is not property_.default))

    @classmethod
    def declarative_from_yaml(cls, constructor, tag_suffix, node):
        """Convert YAML to declarative class instances."""
        try:
            class_ = cls._get_class(tag_suffix)
        except ImportError:
            raise ConstructorError(
                "while constructing a Stone Soup component", node.start_mark,
                "unable to import component {!r}".format(tag_suffix),
                node.start_mark)
        properties = [
            data
            for data in constructor.construct_yaml_omap(node)][0]
        try:
            return class_(**properties)
        except Exception as e:
            raise ConstructorError("while constructing Stone Soup component",
                                   node.start_mark, str(e), node.start_mark)

    @classmethod
    @lru_cache(None)
    def _get_class(cls, tag_suffix):
        tag = cls.tag_prefix + tag_suffix
        classes = [
            subclass
            for subclass in Base.subclasses
            if cls.yaml_tag(subclass) == tag]
        if len(classes) > 1:
            warnings.warn(
                "Multiple possible classes found for YAML tag {!r}".format(
                    tag), UserWarning)
        elif not classes:
            module_name, class_name = tag_suffix.rsplit(".", 1)
            module = import_module("..{}".format(module_name), __name__)
            classes = [getattr(module, class_name, None)]
        if classes[0] is None:
            raise ImportError("Unable to find {!r}".format(tag))
        return classes[0]

    def ndarray_to_yaml(self, representer, node):
        """Convert numpy.ndarray to YAML."""
        if node.ndim > 1:
            array = [self._yaml.seq(row) for row in node.tolist()]
            [seq.fa.set_flow_style() for seq in array]
        else:
            array = node.tolist()
        return representer.represent_sequence(
            "!numpy.ndarray", array)

    @staticmethod
    def ndarray_from_yaml(constructor, node):
        """Convert YAML to numpy.ndarray."""
        return np.array(constructor.construct_sequence(node, deep=True))

    @staticmethod
    def timedelta_to_yaml(representer, node):
        """Convert datetime.timedelta to YAML.

        Value is total number of seconds."""
        return representer.represent_scalar(
            "!datetime.timedelta", str(node.total_seconds()))

    @staticmethod
    def timedelta_from_yaml(constructor, node):
        """Convert YAML to datetime.timedelta.

        Value should be total number of seconds."""
        return datetime.timedelta(
            seconds=float(constructor.construct_scalar(node)))

    @staticmethod
    def path_to_yaml(representer, node):
        """Convert path to YAML.

        Value is total number of seconds."""
        return representer.represent_scalar(
            "!pathlib.Path", str(node))

    @staticmethod
    def path_from_yaml(constructor, node):
        """Convert YAML to datetime.timedelta.

        Value should be total number of seconds."""
        return Path(constructor.construct_scalar(node))



# -*- coding: utf-8 -*-
import pytest
from ruamel.yaml.constructor import ConstructorError

from ..serialise import YAML
from ..base import Property


@pytest.fixture()
def serialised_file():
    return YAML()


def test_declarative(base, serialised_file):
    instance = base(2, "20")
    instance.new_property = True

    serialised_str = serialised_file.dumps(instance)

    assert 'property_c' not in serialised_str  # Default, no need to store

    new_instance = serialised_file.load(serialised_str)
    assert isinstance(new_instance, base)
    assert new_instance.property_a == instance.property_a
    assert new_instance.property_b == instance.property_b
    assert new_instance.property_c == instance.property_c
    with pytest.raises(AttributeError):
        new_instance.new_property


def test_nested_declarative(base, serialised_file):
    nested_instance = base(1, "nested")
    instance = base(2, "primary", nested_instance)

    serialised_str = serialised_file.dumps(instance)

    new_instance = serialised_file.load(serialised_str)
    assert isinstance(new_instance, base)
    assert isinstance(new_instance.property_c, base)
    assert new_instance.property_b == "primary"
    assert new_instance.property_c.property_b == "nested"


def test_duplicate_tag_warning(base, serialised_file):
    class _TestDuplicateBase(base):
        pass

    first_class = _TestDuplicateBase

    class _TestDuplicateBase(base):  # noqa:F801
        pass

    second_class = _TestDuplicateBase

    instance = first_class(2, "20")
    instance.new_property = True

    serialised_str = serialised_file.dumps(instance)

    with pytest.warns(UserWarning):
        new_instance = serialised_file.load(serialised_str)

    assert isinstance(new_instance, (first_class, second_class))


def test_numpy(base, serialised_file):
    import numpy as np

    class _TestNumpy(base):
        property_d = Property(np.ndarray)

    instance = _TestNumpy(1, "two",
                          property_d=np.array([[1, 2], [3, 4], [5, 6]]))

    serialised_str = serialised_file.dumps(instance)

    new_instance = serialised_file.load(serialised_str)
    assert isinstance(new_instance.property_d, np.ndarray)
    assert np.allclose(instance.property_d, new_instance.property_d)


def test_datetime(base, serialised_file):
    import datetime

    class _TestNumpy(base):
        property_d = Property(datetime.timedelta)

    instance = _TestNumpy(1, "two",
                          property_d=datetime.timedelta(seconds=500))

    serialised_str = serialised_file.dumps(instance)

    new_instance = serialised_file.load(serialised_str)
    assert isinstance(new_instance.property_d, datetime.timedelta)
    assert instance.property_d == new_instance.property_d


def test_path(serialised_file):
    import pathlib
    import tempfile

    with tempfile.NamedTemporaryFile() as file:
        path = pathlib.Path(file.name)
        serialised_str = serialised_file.dumps(path)
        assert file.name in serialised_str

        new_path = serialised_file.load(serialised_str)
        assert new_path == path


def test_references(base, serialised_file):
    serialised_str = """
        property_b: &prop_b '20'
        test1: &id001 !stonesoup.tests.conftest.base.%3Clocals%3E._TestBase
            - property_a: 2
            - property_b: *prop_b
        test2: *id001
        """

    conf = serialised_file.load(serialised_str)
    new_instance = conf['test2']
    assert new_instance is conf['test1']
    assert isinstance(new_instance, base)
    assert new_instance.property_a == 2
    assert new_instance.property_b == "20"
    assert new_instance.property_c == base.property_c.default


def test_anchor(base, serialised_file):
    instance = base(2, "20")

    serialised_str = serialised_file.dumps(
        [instance, instance,  {"key": instance}])
    assert '&id001' in serialised_str  # Anchor should be created
    assert '*id001' in serialised_str  # Reference should be created

    new_instances = serialised_file.load(serialised_str)
    assert new_instances[0] is new_instances[1]
    assert new_instances[0] is new_instances[2]['key']


def test_bad_tag(serialised_file):
    # Invalid module
    serialised_str = """
        test1: &id001 !stonesoup.tests.this.does.not.exist
            - property_a: 2
            - property_b: "10"
        """

    with pytest.raises(ConstructorError, match="unable to import component"):
        serialised_file.load(serialised_str)

    # Invalid class in valid module
    serialised_str = """
        test1: &id001 !stonesoup.tests.conftest.nope
            - property_a: 2
            - property_b: "10"
        """

    with pytest.raises(ConstructorError, match="unable to import component"):
        serialised_file.load(serialised_str)


def test_missing_property(base, serialised_file):
    instance = base(2, "20")
    instance.new_property = True

    # Pass over `property_b` line.
    serialised_str = "\n".join(
        line
        for line in serialised_file.dumps(instance).split("\n")
        if 'property_b' not in line)

    assert 'property_b' not in serialised_str  # Default, no need to store

    with pytest.raises(ConstructorError, match="missing a required argument"):
        serialised_file.load(serialised_str)



# -*- coding: utf-8 -*-
import datetime

from ..base import Property
from .base import Type


class TimeRange(Type):
    """TimeRange type

    An object representing a time range between two timestamps.

    Can be used to check if timestamp is within via `in` operator

    Example
    -------
    >>> t0 = datetime.datetime(2018, 1, 1, 14, 00)
    >>> t1 = datetime.datetime(2018, 1, 1, 15, 00)
    >>> time_range = TimeRange(t0, t1)
    >>> test_time = datetime.datetime(2018, 1, 1, 14, 30)
    >>> print(test_time in time_range)
    True
    """

    start_timestamp = Property(datetime.datetime,
                               doc="Start of the time range")
    end_timestamp = Property(datetime.datetime,
                             doc="End of the time range")

    def __init__(self, start_timestamp, end_timestamp, *args, **kwargs):
        if end_timestamp < start_timestamp:
            raise ValueError("start_timestamp must be before end_timestamp")
        super().__init__(start_timestamp, end_timestamp, *args, **kwargs)

    @property
    def duration(self):
        """Duration of the time range"""

        return self.end_timestamp - self.start_timestamp

    def __contains__(self, timestamp):
        """Checks if timestamp is within range

        Parameters
        ----------
        timestamp : datetime.datetime
            Time stamp to check if within range

        Returns
        -------
        bool
            `True` if timestamp within :attr:`start_timestamp` and
            :attr:`end_timestamp` (inclusive)
        """

        return self.start_timestamp <= timestamp <= self.end_timestamp



# -*- coding: utf-8 -*-
import numpy as np

from ...types.particle import Particle
from ..particle import SystematicResampler


def test_systematic_equal():
    particles = [Particle(np.array([[i]]), weight=1/20) for i in range(20)]

    resamlper = SystematicResampler()

    new_particles = resamlper.resample(particles)

    # Particles equal weight, so should end up with similar particles.
    assert all(np.array_equal(np.array([[i]]), new_particle.state_vector)
               for i, new_particle in enumerate(new_particles))


def test_systematic_single():
    particles = [Particle(np.array([[i]]), weight=1 if i == 10 else 0)
                 for i in range(20)]

    resamlper = SystematicResampler()

    new_particles = resamlper.resample(particles)

    # Weight all at single particle at index/vector 10, so all should be
    # at same point
    assert all(np.array_equal(np.array([[10]]), new_particle.state_vector)
               for new_particle in new_particles)


def test_systematic_even():
    particles = [Particle(np.array([[i]]), weight=1/10 if i % 2 == 0 else 0)
                 for i in range(20)]

    resamlper = SystematicResampler()

    new_particles = resamlper.resample(particles)

    # Weight all at even particles, so new should all be at even vector
    assert all(np.array_equal(np.array([[i//2*2]]), new_particle.state_vector)
               for i, new_particle in enumerate(new_particles))



import datetime

from ..tracktotruthmetrics import SIAPMetrics
from ...types.association import TimeRangeAssociation, AssociationSet
from ...types.track import Track
from ...types.groundtruth import GroundTruthPath, GroundTruthState
from ...metricgenerator.manager import SimpleManager
from ...types.time import TimeRange
from ...types.state import State


def test_j_t():
    manager = SimpleManager()
    metric = SIAPMetrics()
    tstart = datetime.datetime.now()
    truths = {GroundTruthPath(states=[
        GroundTruthState([[1]],
                         timestamp=tstart + datetime.timedelta(seconds=i))
        for i in range(5)]),
        GroundTruthPath(states=[
            GroundTruthState([[1]], timestamp=tstart + datetime.timedelta(
                seconds=i))
            for i in range(3)])}
    manager.groundtruth_paths = truths

    assert metric._j_t(manager, tstart + datetime.timedelta(seconds=1)) == 2
    assert metric._j_t(manager, tstart + datetime.timedelta(seconds=4)) == 1


def test_j_sum():
    manager = SimpleManager()
    metric = SIAPMetrics()
    tstart = datetime.datetime.now()
    truths = {GroundTruthPath(states=[
        GroundTruthState([[1]], timestamp=tstart + datetime.timedelta(
            seconds=i))
        for i in range(5)]),
        GroundTruthPath(states=[
            GroundTruthState([[1]], timestamp=tstart + datetime.timedelta(
                seconds=i))
            for i in range(3)])
    }
    manager.groundtruth_paths = truths

    assert metric._j_sum(manager, [tstart + datetime.timedelta(seconds=i)
                                   for i in range(7)]) == 8


def test_jt():
    manager = SimpleManager()
    metric = SIAPMetrics()
    tstart = datetime.datetime.now()
    associations = {TimeRangeAssociation([], time_range=TimeRange(
        start_timestamp=tstart,
        end_timestamp=tstart + datetime.timedelta(seconds=3))),
                    TimeRangeAssociation([], time_range=TimeRange(
                        start_timestamp=tstart,
                        end_timestamp=tstart + datetime.timedelta(seconds=5)))}
    manager.association_set = AssociationSet(associations)

    # test _jt_t
    assert metric._jt_t(manager, tstart) == 2
    assert metric._jt_t(manager, tstart + datetime.timedelta(seconds=5)) == 1

    # test _jt_sum
    assert metric._jt_sum(manager,
                          [tstart + datetime.timedelta(seconds=i)
                           for i in range(5)]) == 9


def test__na():
    manager = SimpleManager()
    metric = SIAPMetrics()
    tstart = datetime.datetime.now()
    truth = GroundTruthPath(states=[
        GroundTruthState([[1]], timestamp=tstart + datetime.timedelta(
            seconds=i))
        for i in range(5)])
    track1 = Track(
        states=[State([[1]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(5)])
    track2 = Track(
        states=[State([[2]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(3)])
    associations = {TimeRangeAssociation({truth, track1}, time_range=TimeRange(
        start_timestamp=tstart,
        end_timestamp=tstart + datetime.timedelta(seconds=4))),
                    TimeRangeAssociation({truth, track2}, time_range=TimeRange(
                        start_timestamp=tstart,
                        end_timestamp=tstart + datetime.timedelta(seconds=2)))}
    manager.tracks = {track1, track2}
    manager.association_set = AssociationSet(associations)

    # Test _na_t
    assert metric._na_t(manager, tstart + datetime.timedelta(seconds=1)) == 2
    assert metric._na_t(manager, tstart + datetime.timedelta(seconds=4)) == 1
    assert metric._na_t(manager, tstart + datetime.timedelta(seconds=7)) == 0

    # Test _na_sum

    assert metric._na_sum(manager, [tstart + datetime.timedelta(seconds=i)
                                    for i in range(4)]) == 7


def test_n():
    manager = SimpleManager()
    metric = SIAPMetrics()
    tstart = datetime.datetime.now()
    track1 = Track(
        states=[State([[1]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(5)])
    track2 = Track(
        states=[State([[2]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(3)])
    track3 = Track(
        states=[State([[3]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(2, 7)])
    manager.tracks = {track1, track2, track3}

    # test _n_t
    assert metric._n_t(manager, tstart + datetime.timedelta(seconds=2)) == 3
    assert metric._n_t(manager, tstart + datetime.timedelta(seconds=4)) == 2
    assert metric._n_t(manager, tstart + datetime.timedelta(seconds=6)) == 1
    assert metric._n_t(manager, tstart + datetime.timedelta(seconds=10)) == 0

    # test _n_sum
    assert metric._n_sum(manager, [tstart + datetime.timedelta(seconds=i)
                                   for i in range(5)]) == 11


def test_tt_j():
    manager = SimpleManager()
    metric = SIAPMetrics()
    tstart = datetime.datetime.now()
    truth = GroundTruthPath(states=[
        GroundTruthState([[1]], timestamp=tstart + datetime.timedelta(
            seconds=i))
        for i in range(20)])
    # Idea is track 1, then no track then 2 then 2 and 3 then 3
    track1 = Track(
        states=[State([[1]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(3)])
    track2 = Track(
        states=[State([[2]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(5, 10)])
    track3 = Track(
        states=[State([[3]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(7, 15)])
    associations = {TimeRangeAssociation({truth, track1}, time_range=TimeRange(
        start_timestamp=tstart,
        end_timestamp=tstart + datetime.timedelta(seconds=2))),
                    TimeRangeAssociation({truth, track2}, time_range=TimeRange(
                        start_timestamp=tstart + datetime.timedelta(seconds=5),
                        end_timestamp=tstart + datetime.timedelta(seconds=9))),
                    TimeRangeAssociation({truth, track3}, time_range=TimeRange(
                        start_timestamp=tstart + datetime.timedelta(seconds=7),
                        end_timestamp=tstart + datetime.timedelta(
                            seconds=14)))}
    manager.tracks = {track1, track2, track3}
    manager.association_set = AssociationSet(associations)

    assert metric._tt_j(manager, truth) == datetime.timedelta(seconds=11)


def test_nu_j():
    manager = SimpleManager()
    metric = SIAPMetrics()
    tstart = datetime.datetime.now()
    truth = GroundTruthPath(states=[
        GroundTruthState([[1]], timestamp=tstart + datetime.timedelta(
            seconds=i))
        for i in range(40)])
    # Single track
    track1 = Track(
        states=[State([[1]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(3)])
    # Overlapping tracks
    track2 = Track(
        states=[State([[2]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(5, 10)])
    track3 = Track(
        states=[State([[3]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(7, 15)])
    track4 = Track(
        states=[State([[4]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(13, 20)])
    # Long track with shorter unnecessary one internal
    track5 = Track(
        states=[State([[5]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(18, 28)])
    track6 = Track(
        states=[State([[6]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(22, 26)])
    # 2 tracks covered by same range as 1
    track7 = Track(
        states=[State([[7]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(30, 40)])
    track8 = Track(
        states=[State([[8]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(30, 35)])
    track9 = Track(
        states=[State([[9]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(35, 40)])

    manager.tracks = {track1, track2, track3, track4, track5, track6, track7,
                      track8, track9}
    manager.groundtruth_paths = {truth}
    associations = {TimeRangeAssociation({truth, track}, time_range=TimeRange(
        start_timestamp=min([state.timestamp for state in track.states]),
        end_timestamp=max([state.timestamp for state in track.states])))
                    for track in manager.tracks}
    manager.association_set = AssociationSet(associations)

    # test nu_j
    # The fewest tracks to cover the whole length should be tracks 1,2,3,5 & 7
    assert metric._nu_j(manager, truth) == 6

    # test _tl_j
    # longest track on truth is track 7
    assert metric._tl_j(manager, truth) == datetime.timedelta(seconds=9)

    # test _r
    assert metric._r(manager) == 5 / 33


def test_t_j():
    metric = SIAPMetrics()

    tstart = datetime.datetime.now()
    truth = GroundTruthPath(states=[
        GroundTruthState([[1]], timestamp=tstart + datetime.timedelta(
            seconds=i))
        for i in range(40)])

    assert metric._t_j(truth) == datetime.timedelta(seconds=39)


def test_compute_metric():
    manager = SimpleManager()
    generator = SIAPMetrics()
    # Create truth, tracks and associations, same as test_nu_j
    tstart = datetime.datetime.now()
    truth = GroundTruthPath(states=[
        GroundTruthState([[1]], timestamp=tstart + datetime.timedelta(
            seconds=i))
        for i in range(40)])
    track1 = Track(
        states=[State([[1]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(3)])
    track2 = Track(
        states=[State([[2]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(5, 10)])
    track3 = Track(
        states=[State([[3]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(7, 15)])
    track4 = Track(
        states=[State([[4]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(13, 20)])
    track5 = Track(
        states=[State([[5]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(18, 28)])
    track6 = Track(
        states=[State([[6]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(22, 26)])
    track7 = Track(
        states=[State([[7]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(30, 40)])
    track8 = Track(
        states=[State([[8]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(30, 35)])
    track9 = Track(
        states=[State([[9]], timestamp=tstart + datetime.timedelta(seconds=i))
                for i in range(35, 40)])
    manager.groundtruth_paths = truth
    manager.tracks = {track1, track2, track3, track4, track5, track6, track7,
                      track8, track9}
    manager.groundtruth_paths = {truth}
    associations = {TimeRangeAssociation({truth, track}, time_range=TimeRange(
        start_timestamp=min([state.timestamp for state in track.states]),
        end_timestamp=max([state.timestamp for state in track.states])))
                    for track in manager.tracks}
    manager.association_set = AssociationSet(associations)

    metrics = generator.compute_metric(manager)
    tend = tstart + datetime.timedelta(seconds=39)

    assert len(metrics) == 5

    c = metrics[0]
    assert c.title == "SIAP C"
    assert c.value == generator._jt_sum(manager,
                                        manager.list_timestamps()) / (
        generator._j_sum(manager, manager.list_timestamps()))
    assert c.time_range.start_timestamp == tstart
    assert c.time_range.end_timestamp == tend
    assert c.generator == generator

    a = metrics[1]
    assert a.title == "SIAP A"
    assert a.value == generator._na_sum(manager,
                                        manager.list_timestamps()) / (
               generator._jt_sum(manager, manager.list_timestamps()))
    assert a.time_range.start_timestamp == tstart
    assert a.time_range.end_timestamp == tend
    assert a.generator == generator

    s = metrics[2]
    assert s.title == "SIAP S"
    assert s.value == sum([generator._n_t(manager, timestamp) -
                           generator._na_t(manager, timestamp)
                           for timestamp in manager.list_timestamps()]) / (
        generator._n_sum(manager, manager.list_timestamps()))
    assert s.time_range.start_timestamp == tstart
    assert s.time_range.end_timestamp == tend
    assert s.generator == generator

    lt = metrics[3]
    assert lt.title == "SIAP LT"
    assert lt.value == 1 / generator._r(manager)
    assert lt.time_range.start_timestamp == tstart
    assert lt.time_range.end_timestamp == tend
    assert lt.generator == generator

    ls = metrics[4]
    assert ls.title == "SIAP LS"
    assert ls.value == sum([generator._tl_j(manager, truth).total_seconds()
                            for truth in manager.groundtruth_paths]) / sum(
        [generator._t_j(truth).total_seconds()
         for truth in manager.groundtruth_paths])
    assert ls.time_range.start_timestamp == tstart
    assert ls.time_range.end_timestamp == tend
    assert ls.generator == generator



from itertools import chain

import numpy as np
from scipy.optimize import linear_sum_assignment

from .base import MetricGenerator
from ..base import Property
from ..models.measurement import MeasurementModel
from ..types.state import State, StateMutableSequence
from ..types.time import TimeRange
from ..types.metric import SingleTimeMetric, TimeRangeMetric


class GOSPAMetric(MetricGenerator):
    """
    Computes the Generalized Optimal SubPattern Assignment (GOPSA) metric
    for two sets of :class:`~.Track` objects. This implementation of GOSPA
    is based on the auction algorithm.

    The GOPSA metric is calculated at each time step in which a
    :class:`~.Track` object is present

    Reference:
        [1] A. S. Rahmathullah, A. F. García-Fernández, L. Svensson,
        Generalized optimal sub-pattern assignment metric, 2016,
        [online] Available: http://arxiv.org/abs/1601.05585.
    """
    p = Property(float, doc="1<=p<infty, exponent.")
    c = Property(float, doc="c>0, cutoff distance.")
    measurement_model_truth = Property(
        MeasurementModel,
        doc="Measurement model which specifies which elements within the"
            "truth state are to be used to calculate distance over")
    measurement_model_track = Property(
        MeasurementModel,
        doc="Measurement model which specifies which elements within"
            "the truth state are to be used to calculate distance over")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alpha = 2

    def compute_metric(self, manager):
        """Compute the metric using the data in the metric manager

        Parameters
        ----------
        manager : :class:`~.MetricManager`
            contains the data to be used to create the metric(s)

        Returns
        -------
        metric : list :class:`~.Metric`
            Containing the metric information. The value of the metric is a
            list of metrics at each timestamp

        """

        metric = self.process_datasets(manager.tracks,
                                       manager.groundtruth_paths)
        return metric

    def process_datasets(self, dataset_1, dataset_2):
        """Process a dataset of point patterns to provide metric over time

        Parameters
        ----------
        dataset_1: object containing :class:`~.state`
        dataset_2: object containing :class:`~.state`

        Returns
        -------
        metrics: :class:`~.Metric`
            Containing the metric at each timestamp in the form of a list of
            :class:`~.Metric` objects
        """

        states_1 = self.extract_states(dataset_1)
        states_2 = self.extract_states(dataset_2)
        return self.compute_over_time(states_1, states_2)

    @staticmethod
    def extract_states(object_with_states):
        """
        Extracts a list of :class:`~states` from a list of (or single) objects
        containing states. This method is defined to handle :class:`~track`,
        :class:`~groundtruthpath` and :class:`~detection` objects

        Parameters
        ----------
        object_with_states: object containing a list of states
            Method of state extraction depends on the type of the object

        Returns
        -------
        : list of :class:`~.State`
        """

        state_list = StateMutableSequence()
        for element in list(object_with_states):
            if isinstance(element, StateMutableSequence):
                state_list.extend(element.states)
            elif isinstance(element, State):
                state_list.append(element)
            else:
                raise ValueError(
                    "{!r} has no state extraction method".format(element))

        return state_list

    def compute_over_time(self, measured_states, truth_states):
        """
        Compute the GOSPA metric at every timestep from a list of measured
        states and truth states.

        Parameters
        ----------

        measured_states: List of states created by a filter
        truth_states: List of truth states to compare against

        Returns
        -------
        metric: :class:`~.TimeRangeMetric` covering the duration that states
        exist for in the parameters. metric.value contains a list of metrics
        for the GOSPA metric at each timestamp
        """

        # Make a list of all the unique timestamps used
        # Make a sorted list of all the unique timestamps used
        timestamps = sorted({
            state.timestamp
            for state in chain(measured_states, truth_states)})

        gospa_metrics = []

        for timestamp in timestamps:
            meas_points = [state
                           for state in measured_states
                           if state.timestamp == timestamp]
            truth_points = [state
                            for state in truth_states
                            if state.timestamp == timestamp]
            metric, truth_to_measured_assignment = self.compute_gospa_metric(
                    meas_points, truth_points)
            gospa_metrics.append(metric)

        # If only one timestamp is present then return a SingleTimeMetric
        if len(timestamps) == 1:
            return gospa_metrics[0]
        else:
            return TimeRangeMetric(
                title='GOSPA Metrics',
                value=gospa_metrics,
                time_range=TimeRange(min(timestamps), max(timestamps)),
                generator=self)

    def compute_assignments(self, cost_matrix, max_iter):
        """Compute assignments using Auction Algorithm.

        Parameters
        ----------
        cost_matrix: Matrix (size mxn) that denotes the cost of assigning
                            mth truth state to each of the n measured states.
        max_iter: Maximum number of iterations to perform

        Returns
        ---------
        truth_to_measured: Vector of size 1xm, which has indices of the
                    measured objects or '-1' if unassigned.
        measured_to_truth: Vector of size 1xn, which has indices of the
                            truth objects or '-1' if unassigned.
        opt_cost: Scalar value of the optimal assignment
        """

        m_truth, n_measured = cost_matrix.shape
        # Index for objects that will be left un-assigned.
        unassigned_idx = -1

        opt_cost = 0.0
        measured_to_truth = -1 * np.ones([1, n_measured], dtype=np.int64)
        truth_to_measured = -1 * np.ones([1, m_truth], dtype=np.int64)

        if m_truth == 1:
            # Corner case 1: if there is only one truth state.
            opt_cost = np.max(cost_matrix)
            max_cost_idx = np.where(cost_matrix == opt_cost)[1]
            truth_to_measured[0, 0] = max_cost_idx[0]
            measured_to_truth[0, truth_to_measured[0, 0]] = 1

            return truth_to_measured, measured_to_truth, opt_cost

        if n_measured == 1:
            # Corner case 1: if there is only one measured state.
            opt_cost = np.max(cost_matrix)
            max_cost_idx = np.where(cost_matrix == opt_cost)[1]
            measured_to_truth[0, 0] = max_cost_idx[0]
            truth_to_measured[0, measured_to_truth[0, 0]] = 1

            return truth_to_measured, measured_to_truth, opt_cost

        swap_dim_flag = False
        epsil = 1. / np.max([m_truth, n_measured])

        if n_measured < m_truth:
            # The implementation only works when
            # m_truth <= n_measured
            # So swap cost matrix
            cost_matrix = cost_matrix.transpose()
            m_truth, n_measured = cost_matrix.shape
            tmp = measured_to_truth
            measured_to_truth = truth_to_measured
            truth_to_measured = tmp
            swap_dim_flag = True

        # Initial cost for each measured state
        c_measured = np.zeros([1, n_measured])
        k_iter = 0

        while not np.all(truth_to_measured != unassigned_idx):
            if k_iter > max_iter:
                # Raise max iterations reached warning.
                break
            for i in range(m_truth):
                if truth_to_measured[0, i] == unassigned_idx:
                    # Unassigned truth object 'i' bids for the best
                    # measured object j_star

                    # Value for each measured object for truth 'i'
                    tmp_mat = cost_matrix[i, :] - c_measured
                    val_i_j = np.sort(tmp_mat[0, :])[::-1]
                    j = np.argsort(tmp_mat[0, :])[::-1]

                    # Best measurement for truth 'i'
                    j_star = j[0]

                    # 1st and 2nd best value for truth 'i'
                    v_i_j_star = val_i_j[0]
                    w_i_j_star = val_i_j[1]
                    # Bid for measured j_star
                    if w_i_j_star != (-1. * np.inf):
                        c_measured[0, j_star] = c_measured[
                            0, j_star] + v_i_j_star - w_i_j_star + epsil
                    else:
                        c_measured[0, j_star] = c_measured[
                            0, j_star] + v_i_j_star + epsil

                    # If j_star is unassigned
                    if measured_to_truth[0, j_star] != unassigned_idx:

                        opt_cost = opt_cost - \
                            cost_matrix[measured_to_truth[0, j_star], j_star]
                        truth_to_measured[0, measured_to_truth[
                            0, j_star]] = unassigned_idx

                    measured_to_truth[0, j_star] = i
                    truth_to_measured[0, i] = j_star

                    # update the cost of new assignment
                    opt_cost = opt_cost + cost_matrix[i, j_star]
            k_iter += 1

        if swap_dim_flag:
            tmp = measured_to_truth
            measured_to_truth = truth_to_measured
            truth_to_measured = tmp

        return truth_to_measured, measured_to_truth, opt_cost

    def compute_cost_matrix(self, track_states, truth_states):
        """
        Creates the cost matrix between two lists of states

        Parameters
        ----------
        track_states: list of states
        truth_states: list of states

        Returns
        ----------
        cost_matrix: Matrix of euclidian distance between each element in each
        list of states
        """

        cost_matrix = np.ones([len(track_states), len(truth_states)]) * self.c

        for i_track, track_state, in enumerate(track_states):
            for i_truth, truth_state in enumerate(truth_states):

                euc_distance = np.linalg.norm(
                    self.measurement_model_track.function(
                        track_state.state_vector, noise=0)
                    - self.measurement_model_truth.function(
                        truth_state.state_vector, noise=0))

                if euc_distance < self.c:
                    cost_matrix[i_track, i_truth] = euc_distance

        return cost_matrix

    def compute_gospa_metric(self, measured_states, truth_states):
        """Computes GOSPA metric between measured and truth states.

        Parameters
        ----------
        measured_states: list of :class:`~.State`
            list of state objects to be assigned to the truth
        truth_states: list of :class:`~.State`
            list of state objects for the truth points

        Returns
        -------
        gospa_metric: Dictionary containing GOSPA metric for alpha = 2.
                      GOSPA metric is divided into four components:
                      1. distance, 2. localisation, 3. missed, and 4. false.
                      Note that distance = (localisation + missed + false)^1/p
        truth_to_measured_assignment: Assignment matrix.
        """
        timestamps = {
            state.timestamp
            for state in chain(truth_states, measured_states)}
        if len(timestamps) != 1:
            raise ValueError(
                'All states must be from the same time to compute GOSPA')

        gospa_metric = {'distance': 0.0,
                        'localisation': 0.0,
                        'missed': 0,
                        'false': 0}
        num_truth_states = len(truth_states)
        num_measured_states = len(measured_states)
        truth_to_measured_assignment = []
        cost_matrix = self.compute_cost_matrix(measured_states, truth_states)
        cost_matrix = cost_matrix.transpose()

        opt_cost = 0.0
        dummy_cost = (self.c ** self.p) / self.alpha

        if num_truth_states == 0:
            # When truth states are empty all measured states are false
            opt_cost = -1.0 * num_measured_states * dummy_cost
            # num_false = opt_cost
        else:
            if num_measured_states == 0:
                # When measured states are empty all truth
                # states are missed
                opt_cost = -1. * num_truth_states * dummy_cost
                if self.alpha == 2:
                    self.missed = opt_cost
            else:
                # Use auction algorithm when both truth_states
                # and measured_states are non-empty
                cost_matrix = -1. * np.power(cost_matrix, self.p)
                truth_to_measured_assignment, measured_to_truth_assignment,\
                    opt_cost_tmp =\
                    self.compute_assignments(cost_matrix,
                                             10 * num_truth_states *
                                             num_measured_states)
                # Now use assignments to compute bids
                for i in range(num_truth_states):
                    if truth_to_measured_assignment[0, i] != -1:
                        opt_cost = opt_cost +\
                            cost_matrix[i, truth_to_measured_assignment[0, i]]

                        if self.alpha == 2:
                            const_assign = \
                                truth_to_measured_assignment[0, i]
                            const_cmp = (-1 * self.c**self.p)

                            gospa_metric['localisation'] = \
                                gospa_metric['localisation'] +\
                                cost_matrix[i, const_assign] *\
                                np.double(
                                    cost_matrix[i, const_assign] > const_cmp)

                            gospa_metric['missed'] = gospa_metric['missed'] -\
                                dummy_cost * np.double(
                                    cost_matrix[i, const_assign] == const_cmp)

                            gospa_metric['false'] = gospa_metric['false'] -\
                                dummy_cost * np.double(
                                    cost_matrix[i, const_assign] == const_cmp)
                    else:
                        opt_cost = opt_cost - dummy_cost
                        if self.alpha == 2:
                            gospa_metric['missed'] = gospa_metric[
                                'missed'] - dummy_cost

                opt_cost = opt_cost - \
                    np.sum(measured_to_truth_assignment == -1) * dummy_cost
                if self.alpha == 2:
                    gospa_metric['false'] = gospa_metric['false'] - \
                        np.sum(measured_to_truth_assignment == -1) * dummy_cost
        gospa_metric['distance'] = np.power((-1. * opt_cost), 1 / self.p)
        gospa_metric['localisation'] = -1. * gospa_metric['localisation']
        gospa_metric['missed'] = -1 * gospa_metric['missed']
        gospa_metric['false'] = -1 * gospa_metric['false']

        single_time_gospa_metric = SingleTimeMetric(
                title='GOSPA Metric', value=gospa_metric,
                timestamp=timestamps.pop(), generator=self)

        return single_time_gospa_metric, truth_to_measured_assignment


class OSPAMetric(GOSPAMetric):
    """
    Computes the Optimal SubPattern Assignment (OPSA) distance [1] for two sets
    of :class:`~.Track` objects. The OSPA distance is measured between two
    point patterns.

    The OPSA metric is calculated at each time step in which a :class:`~.Track`
    object is present

    Reference:
        [1] A Consistent Metric for Performance Evaluation of Multi-Object
        Filters, D. Schuhmacher, B. Vo and B. Vo, IEEE Trans. Signal Processing
        2008
    """

    c = Property(float, doc='Maximum distance for possible association')
    p = Property(float, doc='norm associated to distance')

    def compute_over_time(self, measured_states, truth_states):
        """Compute the OSPA metric at every timestep from a list of measured
        states and truth states

        Parameters
        ----------
        measured_states: list of :class:`~.State`
            Created by a filter
        truth_states: list of :class:`~.State`
            Truth states to compare against

        Returns
        -------
        TimeRangeMetric
            Covering the duration that states exist for in the parameters.
            Metric.value contains a list of metrics for the OSPA distance at
            each timestamp
        """

        # Make a sorted list of all the unique timestamps used
        timestamps = sorted({
            state.timestamp
            for state in chain(measured_states, truth_states)})

        ospa_distances = []

        for timestamp in timestamps:
            meas_points = [state
                           for state in measured_states
                           if state.timestamp == timestamp]
            truth_points = [state
                            for state in truth_states
                            if state.timestamp == timestamp]
            ospa_distances.append(
                self.compute_OSPA_distance(meas_points, truth_points))

        # If only one timestamp is present then return a SingleTimeMetric
        if len(timestamps) == 1:
            return ospa_distances[0]
        else:
            return TimeRangeMetric(
                title='OSPA distances',
                value=ospa_distances,
                time_range=TimeRange(min(timestamps), max(timestamps)),
                generator=self)

    def compute_OSPA_distance(self, track_states, truth_states):
        r"""
        Computes the Optimal SubPattern Assignment (OPSA) metric for a single
        time step between two point patterns. Each point pattern consisting of
        a list of :class:`~.State` objects.

        The function :math:`\bar{d}_{p}^{(c)}` is the OSPA metric of order
        :math:`p` with cut-off :math:`c`. The OSPA metric is defined as:

            .. math::
                \begin{equation*}
                    \bar{d}_{p}^{(c)}({X},{Y}) :=
                    \Biggl( \frac{1}{n}
                    \Bigl({min}_{\substack{
                        \pi\in\Pi_{n}}}
                            \sum_{i=1}^{m}
                                d^{(c)}(x_{i},y_{\pi(i)})^{p}+
                                c^{p}(n-m)\Bigr)
                        \Biggr)^{ \frac{1}{p} }
                \end{equation*}

        Parameters
        ----------
        track_states: list of :class:`~.State`
        truth_states: list of :class:`~.State`

        Returns
        -------

        SingleTimeMetric
            The OSPA distance

        """

        timestamps = {
            state.timestamp
            for state in chain(truth_states, track_states)}
        if len(timestamps) != 1:
            raise ValueError(
                'All states must be from the same time to perform OSPA')

        if not track_states or not truth_states:
            distance = 0
        else:
            cost_matrix = self.compute_cost_matrix(track_states, truth_states)

            # Solve cost matrix with Hungarian/Munkres using
            # scipy.optimize.linear_sum_assignemnt
            # Length of longest set of states
            n = max(len(track_states), len(truth_states))
            row_ind, col_ind = linear_sum_assignment(cost_matrix)
            # Calculate metric
            distance = ((1 / n) * cost_matrix[row_ind, col_ind].sum()) ** (
                        1 / self.p)

        return SingleTimeMetric(title='OSPA distance', value=distance,
                                timestamp=timestamps.pop(), generator=self)



# -*- coding: utf-8 -*-

import numpy as np
from functools import lru_cache

from .base import Predictor
from ..base import Property
from ..functions import gauss2sigma, unscented_transform
from ..types.prediction import GaussianStatePrediction
from ..types.state import State


class KalmanPredictor(Predictor):
    """KalmanPredictor class

    An implementation of a standard Kalman Filter predictor.

    """

    @lru_cache()
    def predict(self, prior, control_input=None, timestamp=None, **kwargs):
        """Kalman Filter state prediction step

        Parameters
        ----------
        prior : :class:`~.GaussianState`
            The prior state
        control_input : :class:`~.State`, optional
            The control input. It will only have an effect if
            :attr:`control_model` is not `None` (the default is `None`)
        timestamp: :class:`datetime.datetime`, optional
            A timestamp signifying when the prediction is performed \
            (the default is `None`)

        Returns
        -------
        : :class:`~.GaussianStatePrediction`
            The predicted state

        """

        # Compute time_interval
        try:
            time_interval = timestamp - prior.timestamp
        except TypeError:
            # TypeError: (timestamp or prior.timestamp) is None
            time_interval = None

        # Transition model parameters
        transition_matrix = self.transition_model.matrix(
            timestamp=timestamp,
            time_interval=time_interval,
            **kwargs)
        transition_noise_covar = self.transition_model.covar(
            timestamp=timestamp,
            time_interval=time_interval,
            **kwargs)

        # Control model parameters
        if self.control_model is None:
            control_matrix = np.zeros(prior.covar.shape)
            contol_noise_covar = np.zeros(prior.covar.shape)
            control_input = State(np.zeros(prior.state_vector.shape))
        else:
            # Extract control matrix
            control_matrix = self.control_model.matrix(
                timestamp=timestamp,
                time_interval=time_interval,
                **kwargs)
            # Extract control noise covariance
            try:
                # covar() is implemented for control_model
                contol_noise_covar = self.control_model.covar(
                    timestamp=timestamp,
                    time_interval=time_interval,
                    **kwargs)
            except AttributeError:
                # covar() is NOT implemented for control_model
                contol_noise_covar = np.zeros(self.control_model.ndim_ctrl)
            if control_input is None:
                control_input = np.zeros((self.control_model.ndim_ctrl, 1))

        # Perform prediction
        prediction_mean, prediction_covar = self.predict_lowlevel(
            prior.mean, prior.covar, transition_matrix,
            transition_noise_covar, control_input.state_vector,
            control_matrix, contol_noise_covar)

        return GaussianStatePrediction(prediction_mean,
                                       prediction_covar,
                                       timestamp)

    @staticmethod
    def predict_lowlevel(x, P, F, Q, u, B, Qu):
        """Low-level Kalman Filter state prediction

        Parameters
        ----------
        x : :class:`numpy.ndarray` of shape (Ns,1)
            The prior state mean
        P : :class:`numpy.ndarray` of shape (Ns,Ns)
            The prior state covariance
        F : :class:`numpy.ndarray` of shape (Ns,Ns)
            The state transition matrix
        Q : :class:`numpy.ndarray` of shape (Ns,Ns)
            The process noise covariance matrix
        u : :class:`numpy.ndarray` of shape (Nu,1)
            The control input
        B : :class:`numpy.ndarray` of shape (Ns,Nu)
            The control gain matrix
        Qu : :class:`numpy.ndarray` of shape (Ns,Ns)
            The control process covariance matrix

        Returns
        -------
        : :class:`numpy.ndarray` of shape (Ns,1)
            The predicted state mean
        : :class:`numpy.ndarray` of shape (Ns,Ns)
            The predicted state covariance
        """

        x_pred = F@x + B@u
        P_pred = F@P@F.T + Q + B@Qu@B.T

        return x_pred, P_pred


class ExtendedKalmanPredictor(KalmanPredictor):
    """ExtendedKalmanPredictor class

    An implementation of an Extended Kalman Filter predictor"""

    @lru_cache()
    def predict(self, prior, control_input=None, timestamp=None, **kwargs):
        """ Extended Kalman Filter state prediction step

        Parameters
        ----------
        prior : :class:`~.GaussianState`
            The prior state
        control_input : :class:`~.State`, optional
            The control input. It will only have an effect if
            :attr:`control_model` is not `None` (the default is `None`)
        timestamp: :class:`datetime.datetime`, optional
            A timestamp signifying when the prediction is performed \
            (the default is `None`)

        Returns
        -------
        : :class:`~.GaussianStatePrediction`
            The predicted state
        """

        # Compute time_interval
        try:
            time_interval = timestamp - prior.timestamp
        except TypeError:
            # TypeError: (timestamp or prior.timestamp) is None
            time_interval = None

        # Transition model parameters
        try:
            # Attempt to extract matrix from a LinearModel
            transition_matrix = self.transition_model.matrix(
                timestamp=timestamp,
                time_interval=time_interval,
                **kwargs)
        except AttributeError:
            # Else read jacobian from a NonLinearModel
            transition_matrix = self.transition_model.jacobian(
                state_vec=prior.state_vector,
                timestamp=timestamp,
                time_interval=time_interval,
                **kwargs)

        def transition_function(x):
            return self.transition_model.function(x, timestamp=timestamp,
                                                  time_interval=time_interval,
                                                  noise=0, **kwargs)

        transition_noise_covar = self.transition_model.covar(
            timestamp=timestamp,
            time_interval=time_interval,
            **kwargs)

        # Control model parameters
        if self.control_model is None:
            control_matrix = np.zeros(prior.covar.shape)
            contol_noise_covar = np.zeros(prior.covar.shape)
            control_input = State(np.zeros(prior.state_vector.shape))
        else:
            # Extract control matrix
            try:
                # Attempt to extract matrix from a LinearModel
                control_matrix = self.control_model.matrix(
                    timestamp=timestamp,
                    time_interval=time_interval,
                    **kwargs)
            except AttributeError:
                # Else read jacobian from a NonLinearModel
                control_matrix = self.control_model.jacobian(
                    timestamp=timestamp,
                    time_interval=time_interval,
                    **kwargs)
            # Extract control noise covariance
            try:
                # covar() is implemented for control_model
                contol_noise_covar = self.control_model.covar(
                    timestamp=timestamp,
                    time_interval=time_interval,
                    **kwargs)
            except AttributeError:
                # covar() is NOT implemented for control_model
                contol_noise_covar = np.zeros((self.control_model.ndim_ctrl,
                                               self.control_model.ndim_ctrl))
            if control_input is None:
                control_input = np.zeros((self.control_model.ndim_ctrl, 1))

        # Perform state prediction
        prediction_mean, prediction_covar = self.predict_lowlevel(
            prior.mean, prior.covar, transition_function, transition_matrix,
            transition_noise_covar, control_input.state_vector,
            control_matrix, contol_noise_covar)

        return GaussianStatePrediction(prediction_mean,
                                       prediction_covar,
                                       timestamp)

    @staticmethod
    def predict_lowlevel(x, P, f, F, Q, u, B, Qu):
        """Low-level Extended Kalman Filter state prediction

        Parameters
        ----------
        x : :class:`numpy.ndarray` of shape (Ns,1)
            The prior state mean
        P : :class:`numpy.ndarray` of shape (Ns,Ns)
            The prior state covariance
        f : function handle
            The (non-linear) transition model function
            Must be of the form "xk = fun(xkm1)"
        F : :class:`numpy.ndarray` of shape (Ns,Ns)
            The state transition/jacobian matrix
        Q : :class:`numpy.ndarray` of shape (Ns,Ns)
            The process noise covariance matrix
        u : :class:`numpy.ndarray` of shape (Nu,1)
            The control input
        B : :class:`numpy.ndarray` of shape (Ns,Nu)
            The control gain matrix
        Qu : :class:`numpy.ndarray` of shape (Ns,Ns)
            The control process covariance matrix

        Returns
        -------
        : :class:`numpy.ndarray` of shape (Ns,1)
            The predicted state mean
        : :class:`numpy.ndarray` of shape (Ns,Ns)
            The predicted state covariance
        """

        x_pred = f(x) + B@u
        P_pred = F@P@F.T + Q + B@Qu@B.T

        return x_pred, P_pred


class UnscentedKalmanPredictor(KalmanPredictor):
    """UnscentedKalmanPredictor class


    An implementation of an Unscented Kalman Filter predictor"""

    alpha = Property(float, default=0.5,
                     doc="Primary sigma point spread scalling parameter.\
                         Typically 0.5.")
    beta = Property(float, default=2,
                    doc="Used to incorporate prior knowledge of the distribution.\
                        If the true distribution is Gaussian, the value of 2\
                        is optimal.")
    kappa = Property(float, default=0,
                     doc="Secondary spread scaling parameter\
                        (default is calculated as 3-Ns)")

    @lru_cache()
    def predict(self, prior, control_input=None, timestamp=None, **kwargs):
        """ Unscented Kalman Filter state prediction step

        Parameters
        ----------
        prior : :class:`~.GaussianState`
            The prior state
        control_input : :class:`~.State`, optional
            The control input. It will only have an effect if
            :attr:`control_model` is not `None` (the default is `None`)
        timestamp: :class:`datetime.datetime`, optional
            A timestamp signifying when the prediction is performed \
            (the default is `None`)

        Returns
        -------
        : :class:`~.GaussianStatePrediction`
            The predicted state
        """

        # Compute time_interval
        try:
            time_interval = timestamp - prior.timestamp
        except TypeError:
            # TypeError: (timestamp or prior.timestamp) is None
            time_interval = None

        def transition_function(x, w=0):
            return self.transition_model.function(x, timestamp=timestamp,
                                                  time_interval=time_interval,
                                                  noise=w, **kwargs)

        transition_noise_covar = self.transition_model.covar(
            timestamp=timestamp,
            time_interval=time_interval,
            **kwargs)

        # Control model parameters
        if self.control_model is None:
            control_matrix = np.zeros(prior.covar.shape)
            contol_noise_covar = np.zeros(prior.covar.shape)
            control_input = State(np.zeros(prior.state_vector.shape))
        else:
            # Extract control matrix
            try:
                # Attempt to extract matrix from a LinearModel
                control_matrix = self.control_model.matrix(
                    timestamp=timestamp,
                    time_interval=time_interval,
                    **kwargs)
            except AttributeError:
                # Else read jacobian from a NonLinearModel
                control_matrix = self.control_model.jacobian(
                    timestamp=timestamp,
                    time_interval=time_interval,
                    **kwargs)
            # Extract control noise covariance
            try:
                # covar() is implemented for control_model
                contol_noise_covar = self.control_model.covar(
                    timestamp=timestamp,
                    time_interval=time_interval,
                    **kwargs)
            except AttributeError:
                # covar() is NOT implemented for control_model
                contol_noise_covar = np.zeros((self.control_model.ndim_ctrl,
                                               self.control_model.ndim_ctrl))
            if control_input is None:
                control_input = np.zeros((self.control_model.ndim_ctrl, 1))

        # Perform state prediction
        prediction_mean, prediction_covar = self.predict_lowlevel(
            prior.mean, prior.covar, transition_function,
            transition_noise_covar, control_input.state_vector,
            control_matrix, contol_noise_covar,
            self.alpha, self.beta, self.kappa)

        return GaussianStatePrediction(prediction_mean,
                                       prediction_covar,
                                       timestamp)

    @staticmethod
    def predict_lowlevel(x, P, f, Q, u, B, Qu, alpha, beta, kappa):
        """Low-level Unscented Kalman Filter state prediction

        Parameters
        ----------
        x : :class:`numpy.ndarray` of shape (Ns,1)
            The prior state mean
        P : :class:`numpy.ndarray` of shape (Ns,Ns)
            The prior state covariance
        f : function handle
            The (non-linear) transition model function
            Must be of the form "xk = fun(xkm1)"
        Q : :class:`numpy.ndarray` of shape (Ns,Ns)
            The process noise covariance matrix
        u : :class:`numpy.ndarray` of shape (Nu,1)
            The control input
        B : :class:`numpy.ndarray` of shape (Ns,Nu)
            The control gain matrix
        Qu : :class:`numpy.ndarray` of shape (Ns,Ns)
            The control process covariance matrix
        alpha : float
            Spread of the sigma points.
        beta : float
            Used to incorporate prior knowledge of the distribution
            2 is optimal if the state is normally distributed.
        kappa : float
            Secondary spread scaling parameter

        Returns
        -------
        : :class:`numpy.ndarray` of shape (Ns,1)
            The predicted state mean
        : :class:`numpy.ndarray` of shape (Ns,Ns)
            The predicted state covariance
        """

        sigma_points, mean_weights, covar_weights = \
            gauss2sigma(x, P, alpha, beta, kappa)

        x_pred, P_pred, _, _, _, _ = unscented_transform(sigma_points,
                                                         mean_weights,
                                                         covar_weights,
                                                         f, covar_noise=Q)

        return x_pred, P_pred



# -*- coding: utf-8 -*-
import datetime

import numpy as np

from ..base import Property
from ..models.measurement import MeasurementModel
from ..models.transition import TransitionModel
from ..reader import GroundTruthReader
from ..types.detection import TrueDetection, Clutter
from ..types.groundtruth import GroundTruthPath, GroundTruthState
from ..types.numeric import Probability
from ..types.state import GaussianState, State
from .base import DetectionSimulator, GroundTruthSimulator


class SingleTargetGroundTruthSimulator(GroundTruthSimulator):
    """Target simulator that produces a single target"""
    transition_model = Property(
        TransitionModel, doc="Transition Model used as propagator for track.")
    initial_state = Property(
        State,
        doc="Initial state to use to generate ground truth")
    timestep = Property(
        datetime.timedelta,
        default=datetime.timedelta(seconds=1),
        doc="Time step between each state. Default one second.")
    number_steps = Property(
        int, default=100, doc="Number of time steps to run for")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._groundtruth_paths = set()

    @property
    def groundtruth_paths(self):
        return self._groundtruth_paths.copy()

    def groundtruth_paths_gen(self):
        self._groundtruth_paths = set()
        time = self.initial_state.timestamp or datetime.datetime.now()

        gttrack = GroundTruthPath([
            GroundTruthState(self.initial_state.state_vector, timestamp=time)])
        self._groundtruth_paths.add(gttrack)
        yield time, self.groundtruth_paths

        for _ in range(self.number_steps - 1):
            time += self.timestep
            # Move track forward
            trans_state_vector = self.transition_model.function(
                gttrack[-1].state_vector,
                time_interval=self.timestep)
            gttrack.append(GroundTruthState(
                trans_state_vector, timestamp=time))

            yield time, self.groundtruth_paths


class MultiTargetGroundTruthSimulator(SingleTargetGroundTruthSimulator):
    """Target simulator that produces multiple targets.

    Targets are created and destroyed randomly, as defined by the birth rate
    and death probability."""
    initial_state = Property(
        GaussianState,
        doc="Initial state to use to generate states")
    birth_rate = Property(
        float, default=1.0, doc="Rate at which tracks are born. Expected "
        "number of occurrences (λ) in Poisson distribution. Default 1.0.")
    death_probability = Property(
        Probability, default=0.1,
        doc="Probability of track dying in each time step. Default 0.1.")

    def groundtruth_paths_gen(self):
        self._groundtruth_paths = set()
        time = self.initial_state.timestamp or datetime.datetime.now()

        for _ in range(self.number_steps):
            # Random drop tracks
            self._groundtruth_paths.difference_update(
                gttrack
                for gttrack in self.groundtruth_paths
                if np.random.rand() <= self.death_probability)

            # Move tracks forward
            for gttrack in self.groundtruth_paths:
                trans_state_vector = self.transition_model.function(
                    gttrack[-1].state_vector,
                    time_interval=self.timestep)
                gttrack.append(GroundTruthState(
                    trans_state_vector, timestamp=time))

            # Random create
            for _ in range(np.random.poisson(self.birth_rate)):
                gttrack = GroundTruthPath()
                gttrack.append(GroundTruthState(
                    self.initial_state.state_vector +
                    np.sqrt(self.initial_state.covar) @
                    np.random.randn(self.initial_state.ndim, 1),
                    timestamp=time))
                self._groundtruth_paths.add(gttrack)

            yield time, self.groundtruth_paths
            time += self.timestep


class SimpleDetectionSimulator(DetectionSimulator):
    """A simple detection simulator.

    Parameters
    ----------
    groundtruth : GroundTruthReader
        Source of ground truth tracks used to generate detections for.
    measurement_model : MeasurementModel
        Measurement model used in generating detections.
    """
    groundtruth = Property(GroundTruthReader)
    measurement_model = Property(MeasurementModel)
    meas_range = Property(np.ndarray)
    detection_probability = Property(Probability, default=0.9)
    clutter_rate = Property(float, default=2.0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.real_detections = set()
        self.clutter_detections = set()

    @property
    def detections(self):
        return self.real_detections | self.clutter_detections

    @property
    def clutter_spatial_density(self):
        """returns the clutter spatial density of the measurement space - num
        clutter detections per unit volume per timestep"""
        meas_volume = float(1)
        for dim in range(self.meas_range.ndim):
            meas_volume *= abs(float(self.meas_range[dim][0] *
                                     self.meas_range[dim][-1]))
        return self.clutter_rate/meas_volume

    def detections_gen(self):
        for time, tracks in self.groundtruth.groundtruth_paths_gen():
            self.real_detections.clear()
            self.clutter_detections.clear()

            for track in tracks:
                if np.random.rand() < self.detection_probability:
                    detection = TrueDetection(
                        self.measurement_model.function(
                            track[-1].state_vector),
                        timestamp=track[-1].timestamp,
                        groundtruth_path=track)
                    detection.clutter = False
                    self.real_detections.add(detection)

            # generate clutter
            for _ in range(np.random.poisson(self.clutter_rate)):
                detection = Clutter(
                    np.random.rand(self.measurement_model.ndim_meas, 1) *
                    np.diff(self.meas_range) + self.meas_range[:, :1],
                    timestamp=time)
                self.clutter_detections.add(detection)

            yield time, self.detections



# coding: utf-8
import datetime

import numpy as np

from ..linear import (
    LinearGaussianTimeInvariantTransitionModel, ConstantVelocity,
    CombinedLinearGaussianTransitionModel)


def test_combined():
    F = 3*np.eye(3)
    Q = 3*np.eye(3)
    model_1 = LinearGaussianTimeInvariantTransitionModel(
        transition_matrix=F, covariance_matrix=Q)
    model_2 = ConstantVelocity(noise_diff_coeff=3)
    model_3 = ConstantVelocity(noise_diff_coeff=1)
    model_4 = ConstantVelocity(noise_diff_coeff=1)

    DIM = 9

    combined_model = CombinedLinearGaussianTransitionModel(
        [model_1, model_2, model_3, model_4])
    t_delta = datetime.timedelta(0, 3)

    x_prior = np.ones([DIM, 1])
    x_post = np.ones([DIM, 1])

    assert DIM == combined_model.ndim_state
    assert (DIM, DIM) == combined_model.matrix(time_interval=t_delta).shape
    assert (DIM, DIM) == combined_model.covar(time_interval=t_delta).shape
    assert (DIM, 1) == combined_model.function(
        x_prior, noise=np.random.randn(DIM, 1), time_interval=t_delta).shape
    assert (DIM, 1) == combined_model.rvs(time_interval=t_delta).shape
    assert isinstance(
        combined_model.pdf(x_post, x_prior, time_interval=t_delta), np.float64)



# -*- coding: utf-8 -*-
import numpy as np

from ..particle import Particle


def test_particle():
    particle1 = Particle(np.array([[0]]), weight=0.1)

    assert np.array_equal(particle1.state_vector, np.array([[0]]))
    assert particle1.weight == 0.1

    particle2 = Particle(np.array([[0]]), weight=0.1, parent=particle1)

    assert particle2.parent is particle1

    particle3 = Particle(np.array([[0]]), weight=0.1, parent=particle2)

    assert particle3.parent is particle2
    assert particle3.parent.parent is None



# -*- coding: utf-8 -*-
import numpy as np


class StateVector(np.ndarray):
    """State vector wrapper for :class:`numpy.ndarray`

    This class returns a view to a :class:`numpy.ndarray`, but ensures that
    its initialised at a *Nx1* vector. It's called same as to
    :func:`numpy.asarray`.
    """

    def __new__(cls, *args, **kwargs):
        array = np.asarray(*args, **kwargs)
        if not (array.ndim == 2 and array.shape[1] == 1):
            raise ValueError(
                "state vector shape should be Nx1 dimensions: got {}".format(
                    array.shape))
        return array.view(cls)

    def __array_wrap__(self, array):
        return np.asarray(array)

    def __matmul__(self, other):
        out = np.matmul(np.asfarray(self), np.asfarray(other))
        return out.view(type=type(self))

    def __rmatmul__(self, other):
        out = np.matmul(np.asfarray(other), np.asfarray(self))
        return out.view(type=type(other))


class CovarianceMatrix(np.ndarray):
    """Covariance matrix wrapper for :class:`numpy.ndarray`.

    This class returns a view to a :class:`numpy.ndarray`, but ensures that
    its initialised at a *NxN* matrix. It's called similar to
    :func:`numpy.asarray`.
    """

    def __new__(cls, *args, **kwargs):
        array = np.asarray(*args, **kwargs)
        if not array.ndim == 2:
            raise ValueError("Covariance should have ndim of 2: got {}"
                             "".format(array.ndim))
        return array.view(cls)

    def __array_wrap__(self, array):
        return np.asarray(array)

    def __matmul__(self, other):
        out = np.matmul(self, np.asfarray(other))
        return out.view(type=type(self))

    def __rmatmul__(self, other):
        out = np.matmul(np.asfarray(other), self)
        return out.view(type=type(other))



# -*- coding: utf-8 -*-
from abc import abstractmethod

from ..base import Model


class TransitionModel(Model):
    """Transition Model base class"""

    @property
    @abstractmethod
    def ndim_state(self):
        """Number of state dimensions"""
        pass



# -*- coding: utf-8 -*-
from collections.abc import Sized, Iterable, Container

from ..base import Property
from ..types import Type
from ..types.detection import Detection
from ..types.hypothesis import SingleHypothesis
from ..types.prediction import Prediction
from .numeric import Probability
from .detection import MissedDetection


class MultipleHypothesis(Type, Sized, Iterable, Container):
    """Multiple Hypothesis base type

    A Multiple Hypothesis is a container to store a collection of hypotheses.
    """

    single_hypotheses = Property(
        [SingleHypothesis], default=None,
        doc="The initial list of :class:`~.SingleHypothesis`. Default `None`"
            "which initialises with empty list.")
    normalise = Property(
        bool, default=False,
        doc="Normalise probabilities of :class:`~.SingleHypothesis`. Default"
            "is `False`.")
    total_weight = Property(
        float, default=1,
        doc="When normalising, weights will sum to this. Default is 1.")

    def __init__(self, single_hypotheses=None, normalise=False, *args,
                 **kwargs):
        if single_hypotheses is None:
            single_hypotheses = []

        if any(not isinstance(hypothesis, SingleHypothesis)
               for hypothesis in single_hypotheses):
            raise ValueError("Cannot form MultipleHypothesis out of "
                             "non-SingleHypothesis inputs!")

        super().__init__(single_hypotheses, normalise, *args, **kwargs)

        # normalise the weights of 'single_hypotheses', if indicated
        if self.normalise:
            self.normalise_probabilities()

    def __len__(self):
        return self.single_hypotheses.__len__()

    def __contains__(self, index):
        # check if 'single_hypotheses' contains any SingleHypotheses with
        # Detection matching 'index'
        if isinstance(index, Detection):
            for hypothesis in self.single_hypotheses:
                if hypothesis.measurement is index:
                    return True
            return False

        # check if 'single_hypotheses' contains any SingleHypotheses with
        # Prediction matching 'index'
        if isinstance(index, Prediction):
            for hypothesis in self.single_hypotheses:
                if hypothesis.prediction is index:
                    return True
            return False

        # check if 'single_hypotheses' contains any SingleHypotheses
        # matching 'index'
        if isinstance(index, SingleHypothesis):
            return index in self.single_hypotheses

    def __iter__(self):
        for hypothesis in self.single_hypotheses:
            yield hypothesis

    def __getitem__(self, index):

        # retrieve SingleHypothesis by array index
        if isinstance(index, int):
            return self.single_hypotheses[index]

        # retrieve SingleHypothesis by measurement
        if isinstance(index, Detection):
            for hypothesis in self.single_hypotheses:
                if hypothesis.measurement is index:
                    return hypothesis
            return None

        # retrieve SingleHypothesis by prediction
        if isinstance(index, Prediction):
            for hypothesis in self.single_hypotheses:
                if hypothesis.prediction is index:
                    return hypothesis
            return None

    def normalise_probabilities(self, total_weight=None):
        if total_weight is None:
            total_weight = self.total_weight

        # verify that SingleHypotheses composing this MultipleHypothesis
        # all have Probabilities
        if any(not hasattr(hypothesis, 'probability')
               for hypothesis in self.single_hypotheses):
            raise ValueError("MultipleHypothesis not composed of Probability"
                             " hypotheses!")

        sum_weights = Probability.sum(
            hypothesis.probability for hypothesis in self.single_hypotheses)

        for hypothesis in self.single_hypotheses:
            hypothesis.probability =\
                (hypothesis.probability * total_weight)/sum_weights

    def get_missed_detection_probability(self):
        for hypothesis in self.single_hypotheses:
            if isinstance(hypothesis.measurement, MissedDetection):
                if hasattr(hypothesis, 'probability'):
                    return hypothesis.probability
        return None



# -*- coding: utf-8 -*-
import datetime

import pytest

from ...types.array import StateVector
from ...types.detection import Detection
from ...types.groundtruth import GroundTruthState, GroundTruthPath
from ...types.state import State
from ...types.track import Track


@pytest.fixture()
def detection_reader():
    class TestDetectionReader():
        def __init__(self):
            self.detections = set()

        def detections_gen(self):
            time = datetime.datetime(2018, 1, 1, 14)
            state_vector = StateVector([[0]])
            for i in range(3):
                self.detections = {
                    Detection(state_vector + i,  timestamp=time)
                    for _ in range(i)}
                yield time, self.detections
                time += datetime.timedelta(minutes=1)
    return TestDetectionReader()


@pytest.fixture()
def groundtruth_reader():
    class TestGroundTruthReader():
        def __init__(self):
            self.groundtruth_paths = set()

        def groundtruth_paths_gen(self):
            time = datetime.datetime(2018, 1, 1, 14)
            state_vector = StateVector([[0]])
            for i in range(3):
                self.groundtruth_paths = {
                    GroundTruthPath(
                        [GroundTruthState(
                            state_vector + i + 10*j, timestamp=time)
                         for j in range(i)])
                    for _ in range(i)}
                yield time, self.groundtruth_paths
                time += datetime.timedelta(minutes=1)
    return TestGroundTruthReader()


@pytest.fixture()
def tracker():
    class TestTracker():
        def __init__(self):
            self.tracks = set()

        def tracks_gen(self):
            time = datetime.datetime(2018, 1, 1, 14)
            state_vector = StateVector([[0]])
            for i in range(2):
                self.tracks = {
                    Track(
                        [State(
                            state_vector + i + 10*j, timestamp=time)
                            for j in range(i)],
                        str(k))
                    for k in range(i)}
                yield time, self.tracks
                time += datetime.timedelta(minutes=1)
    return TestTracker()



# -*- coding: utf-8 -*-
from ..base import Property
from .groundtruth import GroundTruthPath
from .state import State, GaussianState, StateVector
from ..models.measurement import MeasurementModel


class Detection(State):
    """Detection type"""
    measurement_model = Property(MeasurementModel, default=None,
                                 doc="The measurement model used to generate the detection\
                         (the default is ``None``)")

    metadata = Property(dict, default=None,
                        doc='Dictionary of metadata items for Detections.')

    def __init__(self, state_vector, *args, **kwargs):
        super().__init__(state_vector, *args, **kwargs)
        if self.metadata is None:
            self.metadata = {}


class GaussianDetection(Detection, GaussianState):
    """GaussianDetection type"""


class Clutter(Detection):
    """Clutter type for detections classed as clutter

    This is same as :class:`~.Detection`, but can be used to identify clutter
    for metrics and analysis purposes.
    """


class TrueDetection(Detection):
    """TrueDetection type for detections that come from ground truth

    This is same as :class:`~.Detection`, but can be used to identify true
    detections for metrics and analysis purposes.
    """

    groundtruth_path = Property(
        GroundTruthPath,
        doc="Ground truth path that this detection came from")


class MissedDetection(Detection):
    """Detection type for a missed detection

    This is same as :class:`~.Detection`, but it is used in
    MultipleHypothesis to indicate the null hypothesis (no
    detections are associated with the specified track).
    """

    state_vector = Property(
        StateVector, default=None,
        doc="State vector. Default `None`.")

    def __init__(self, state_vector=None, *args, **kwargs):
        super().__init__(state_vector, *args, **kwargs)

    def __bool__(self):
        return False



# -*- coding: utf-8 -*-
from ..base import Base, Property
from ..tracker import Tracker
from ..metricgenerator import MetricGenerator


class Writer(Base):
    """Writer base class"""


class MetricsWriter(Writer):
    """Metrics Writer base class.

    Writes out metrics to some form of storage for analysis.
    """

    metric_generator = Property(
        MetricGenerator, doc="Source of metric to be written out")


class TrackWriter(Writer):
    """Track Writer base class.

    Writes out tracks to some form of storage for analysis.
    """

    tracker = Property(Tracker, doc="Source of tracks to be written out")



# -*- coding: utf-8 -*-
from ..base import Property
from ..reader import DetectionReader, SensorDataReader


class Detector(DetectionReader):
    """Detector base class

    A Detector processes :class:`~.SensorData` to generate :class:`~.Detection`
    data.
    """

    sensor = Property(SensorDataReader, doc="Source of sensor data")



# -*- coding: utf-8 -*-
import pytest
import numpy as np


@pytest.fixture()
def transition_model():
    class TestTransitionModel:
        def function(self, state_vector, time_interval):
            return state_vector + time_interval.total_seconds()
    return TestTransitionModel()


@pytest.fixture()
def measurement_model():
    class TestMeasurementModel:
        ndim_state = 4
        ndim_meas = 2

        @staticmethod
        def function(state_vector):
            matrix = np.array([[1, 0, 0, 0], [0, 0, 1, 0]])
            return matrix @ state_vector
    return TestMeasurementModel()



# -*- coding: utf-8 -*-
import datetime

import numpy as np
import pytest

from ..numeric import Probability
from ..particle import Particle
from ..state import State, GaussianState, ParticleState, \
    StateMutableSequence, WeightedGaussianState


def test_state():
    with pytest.raises(TypeError):
        State()

    # Test state initiation without timestamp
    state_vector = np.array([[0], [1]])
    state = State(state_vector)
    assert np.array_equal(state.state_vector, state_vector)

    # Test state initiation with timestamp
    timestamp = datetime.datetime.now()
    state = State(state_vector, timestamp=timestamp)
    assert state.timestamp == timestamp


def test_state_invalid_vector():
    with pytest.raises(ValueError):
        State(np.array([1, 2, 3, 4]))


def test_gaussianstate():
    """ GaussianState Type test """

    with pytest.raises(TypeError):
        GaussianState()

    mean = np.array([[-1.8513], [0.9994], [0], [0]]) * 1e4
    covar = np.array([[2.2128, 0, 0, 0],
                      [0.0002, 2.2130, 0, 0],
                      [0.3897, -0.00004, 0.0128, 0],
                      [0, 0.3897, 0.0013, 0.0135]]) * 1e3
    timestamp = datetime.datetime.now()

    # Test state initiation without timestamp
    state = GaussianState(mean, covar)
    assert(np.array_equal(mean, state.mean))
    assert(np.array_equal(covar, state.covar))
    assert(state.ndim == mean.shape[0])
    assert(state.timestamp is None)

    # Test state initiation with timestamp
    state = GaussianState(mean, covar, timestamp)
    assert(np.array_equal(mean, state.mean))
    assert(np.array_equal(covar, state.covar))
    assert(state.ndim == mean.shape[0])
    assert(state.timestamp == timestamp)


def test_gaussianstate_invalid_covar():
    mean = np.array([[1], [2], [3], [4]])  # 4D
    covar = np.diag([1, 2, 3])  # 3D
    with pytest.raises(ValueError):
        GaussianState(mean, covar)


def test_weighted_gaussian_state():
    mean = np.array([[1], [2], [3], [4]])  # 4D
    covar = np.diag([1, 2, 3])  # 3D
    weight = 0.3
    with pytest.raises(ValueError):
        a = WeightedGaussianState(mean, covar, weight)
        assert a.weight == weight


def test_particlestate():
    with pytest.raises(TypeError):
        ParticleState()

    # 1D
    num_particles = 10
    state_vector1 = np.array([[0]])
    state_vector2 = np.array([[100]])
    weight = Probability(1/num_particles)
    particles = []
    particles.extend(Particle(
        state_vector1, weight=weight) for _ in range(num_particles//2))
    particles.extend(Particle(
        state_vector2, weight=weight) for _ in range(num_particles//2))

    # Test state without timestamp
    state = ParticleState(particles)
    assert np.allclose(state.state_vector, np.array([[50]]))
    assert np.allclose(state.covar, np.array([[2500]]))

    # Test state with timestamp
    timestamp = datetime.datetime.now()
    state = ParticleState(particles, timestamp=timestamp)
    assert np.allclose(state.state_vector, np.array([[50]]))
    assert np.allclose(state.covar, np.array([[2500]]))
    assert state.timestamp == timestamp

    # 2D
    state_vector1 = np.array([[0], [0]])
    state_vector2 = np.array([[100], [200]])
    particles = []
    particles.extend(Particle(
        state_vector1, weight=weight) for _ in range(num_particles//2))
    particles.extend(Particle(
        state_vector2, weight=weight) for _ in range(num_particles//2))

    state = ParticleState(particles)
    assert np.allclose(state.state_vector, np.array([[50], [100]]))
    assert np.allclose(state.covar, np.array([[2500, 5000], [5000, 10000]]))


def test_particlestate_weighted():
    num_particles = 10

    # Half particles at high weight at 0
    state_vector1 = np.array([[0]])
    weight1 = Probability(0.75 / (num_particles / 2))

    # Other half of particles low weight at 100
    state_vector2 = np.array([[100]])
    weight2 = Probability(0.25 / (num_particles / 2))

    particles = []
    particles.extend(Particle(
        state_vector1, weight=weight1) for _ in range(num_particles//2))
    particles.extend(Particle(
        state_vector2, weight=weight2) for _ in range(num_particles//2))

    # Check particles sum to 1 still
    assert pytest.approx(1) == sum(particle.weight for particle in particles)

    # Test state vector is now weighted towards 0 from 50 (non-weighted mean)
    state = ParticleState(particles)
    assert np.allclose(state.state_vector, np.array([[25]]))
    assert np.allclose(state.covar, np.array([[1875]]))


def test_state_mutable_sequence_state():
    state_vector = np.array([[0]])
    timestamp = datetime.datetime(2018, 1, 1, 14)
    delta = datetime.timedelta(minutes=1)
    sequence = StateMutableSequence(
        [State(state_vector, timestamp=timestamp+delta*n)
         for n in range(10)])

    assert sequence.state is sequence.states[-1]
    assert np.array_equal(sequence.state_vector, state_vector)
    assert sequence.timestamp == timestamp + delta*9

    del sequence[-1]
    assert sequence.timestamp == timestamp + delta*8


def test_state_mutable_sequence_slice():
    state_vector = np.array([[0]])
    timestamp = datetime.datetime(2018, 1, 1, 14)
    delta = datetime.timedelta(minutes=1)
    sequence = StateMutableSequence(
        [State(state_vector, timestamp=timestamp+delta*n)
         for n in range(10)])

    assert isinstance(sequence[timestamp:], StateMutableSequence)
    assert isinstance(sequence[5:], StateMutableSequence)
    assert isinstance(sequence[timestamp], State)
    assert isinstance(sequence[5], State)

    assert len(sequence[timestamp:]) == 10
    assert len(sequence[:timestamp]) == 0
    assert len(sequence[timestamp+delta*5:]) == 5
    assert len(sequence[:timestamp+delta*5]) == 5
    assert len(sequence[timestamp+delta*4:timestamp+delta*6]) == 2
    assert len(sequence[timestamp+delta*2:timestamp+delta*8:3]) == 2
    assert len(sequence[timestamp+delta*1:][:timestamp+delta*2]) == 1

    assert sequence[timestamp] == sequence.states[0]

    with pytest.raises(TypeError):
        sequence[timestamp:1]

    with pytest.raises(IndexError):
        sequence[timestamp-delta]



# -*- coding: utf-8 -*-
import datetime
from ..simple import SingleTargetTracker, MultiTargetTracker, \
    MultiTargetMixtureTracker


def test_single_target_tracker(
        initiator, deleter, detector, data_associator, updater):
    tracker = SingleTargetTracker(
        initiator, deleter, detector, data_associator, updater)

    previous_time = datetime.datetime(2018, 1, 1, 13, 59)
    total_tracks = set()
    for time, tracks in tracker.tracks_gen():
        assert time == previous_time + datetime.timedelta(minutes=1)
        assert len(tracks) <= 1  # Shouldn't have more than one track
        for track in tracks:
            assert len(track.states) <= 10  # Deleter should delete these
        total_tracks |= tracks

        previous_time = time

    assert len(total_tracks) >= 2  # Should of had at least 2 over all steps


def test_multi_target_tracker(
        initiator, deleter, detector, data_associator, updater):
    tracker = MultiTargetTracker(
        initiator, deleter, detector, data_associator, updater)

    previous_time = datetime.datetime(2018, 1, 1, 13, 59)
    max_tracks = 0
    total_tracks = set()
    for time, tracks in tracker.tracks_gen():
        assert time == previous_time + datetime.timedelta(minutes=1)
        assert len(tracks) <= 3  # Shouldn't have more than three tracks
        for track in tracks:
            assert len(track.states) <= 10  # Deleter should delete these
        if len(tracks) == 3:
            sorted_vectors = sorted(track.state_vector for track in tracks)
            assert sorted_vectors[0] + 10 == sorted_vectors[1]
            assert sorted_vectors[1] + 10 == sorted_vectors[2]

        max_tracks = max(max_tracks, len(tracks))
        total_tracks |= tracks

        previous_time = time

    assert max_tracks >= 3  # Should of had at least 3 tracks in single step
    assert len(total_tracks) >= 6  # Should of had at least 6 over all steps


def test_multi_target_mixture_tracker(
        initiator, deleter, detector, data_mixture_associator, updater):
    tracker = MultiTargetMixtureTracker(
        initiator, deleter, detector, data_mixture_associator, updater)

    previous_time = datetime.datetime(2018, 1, 1, 13, 59)
    max_tracks = 0
    total_tracks = set()
    for time, tracks in tracker.tracks_gen():
        assert time == previous_time + datetime.timedelta(minutes=1)
        # assert len(tracks) <= 3  # Shouldn't have more than three tracks
        for track in tracks:
            assert len(track.states) <= 10  # Deleter should delete these
        # if len(tracks) == 3:
        #    sorted_vectors = sorted(track.state_vector for track in tracks)
        #    assert sorted_vectors[0] + 10 == sorted_vectors[1]
        #    assert sorted_vectors[1] + 10 == sorted_vectors[2]

        max_tracks = max(max_tracks, len(tracks))
        total_tracks |= tracks

        previous_time = time

    assert max_tracks >= 3  # Should of had at least 3 tracks in single step
    assert len(total_tracks) >= 6  # Should of had at least 6 over all steps



# -*- coding: utf-8 -*-
"""Contains collection of time based deleters"""
from datetime import timedelta

from ..base import Property
from ..types.update import Update
from .base import Deleter


class UpdateTimeStepsDeleter(Deleter):
    """Update Time based deleter

    Identify tracks for deletion which an :class:`~.Update` has occurred in
    last :attr:`time_steps_since_update`.
    """

    time_steps_since_update = Property(
        int, doc="Maximum time steps since last update")

    def check_for_deletion(self, track, **kwargs):
        """Delete track without update with measurements within time steps

        Parameters
        ----------
        track : Track
            Track to check for deletion

        Returns
        -------
        bool
            `True` if track has an :class:`~.Update` with measurements within
            time steps; `False` otherwise.
        """
        return not any(isinstance(state, Update) and state.hypothesis
                       for state in track[-self.time_steps_since_update:])


class UpdateTimeDeleter(Deleter):
    """Update Time based deleter

    Identify tracks for deletion which time of last :class:`~.Update` with
    measurements is greater than :attr:`time_since_update`.
    """

    time_since_update = Property(
        timedelta, doc="Maximum time since last update")

    def check_for_deletion(self, track, timestamp=None, **kwargs):
        """Delete track based on time of last update with measurements

        Parameters
        ----------
        track : Track
            Track to check for deletion
        timestamp : datetime.datetime, optional
            Timestamp to calculate deletion time from. Default `None`, where
            the track's latest timestamp will be used.

        Returns
        -------
        bool
            `True` if track has an :class:`~.Update` with measurements within
            time; `False` otherwise.
        """
        if timestamp is None:
            timestamp = track.timestamp
        return not any(isinstance(state, Update) and state.hypothesis
                       for state in track[timestamp - self.time_since_update:])



# -*- coding: utf-8 -*-
import pytest

from ...types.detection import MissedDetection
from ...types.hypothesis import SingleDistanceHypothesis
from ...types.prediction import (
    GaussianMeasurementPrediction, GaussianStatePrediction)
from ...hypothesiser.probability import PDAHypothesiser


@pytest.fixture()
def hypothesiser():
    class TestGaussianHypothesiser:
        def hypothesise(self, track, detections, timestamp):
            hypotheses = list()
            for detection in detections:
                prediction = GaussianStatePrediction(track.state_vector + 1,
                                                     track.covar * 2,
                                                     detection.timestamp)
                measurement_prediction =\
                    GaussianMeasurementPrediction(prediction.state_vector,
                                                  prediction.covar,
                                                  prediction.timestamp)
                distance = abs(track.state_vector - detection.state_vector)

                hypotheses.append(SingleDistanceHypothesis(
                    prediction, detection, distance, measurement_prediction))

            prediction = GaussianStatePrediction(track.state_vector + 1,
                                                 track.covar * 2, timestamp)
            hypotheses.append(
                SingleDistanceHypothesis(
                    prediction, MissedDetection(timestamp=timestamp), 10))
            return hypotheses
    return TestGaussianHypothesiser()


@pytest.fixture()
def probability_predictor():
    class TestGaussianPredictor:
        def predict(self, prior, control_input=None, timestamp=None, **kwargs):
            return GaussianStatePrediction(prior.state_vector + 1,
                                           prior.covar * 2, timestamp)
    return TestGaussianPredictor()


@pytest.fixture()
def probability_updater():
    class TestGaussianUpdater:
        def get_measurement_prediction(self, state_prediction,
                                       measurement_model=None, **kwargs):
            return GaussianMeasurementPrediction(state_prediction.state_vector,
                                                 state_prediction.covar,
                                                 state_prediction.timestamp)
    return TestGaussianUpdater()


@pytest.fixture()
def probability_hypothesiser(probability_predictor, probability_updater):

    return PDAHypothesiser(probability_predictor, probability_updater,
                           clutter_spatial_density=1.2e-2,
                           prob_detect=0.9, prob_gate=0.99)



# -*- coding: utf-8 -*-
from abc import abstractmethod

from ..base import Base


class Tracker(Base):
    """Tracker base class"""

    @property
    @abstractmethod
    def tracks(self):
        """The tracks at the current time step.

        This is the set of tracks last returned by the
        :meth:`tracks_gen` generator, to allow other components, like
        metrics, to access the data.
        """
        raise NotImplementedError

    @abstractmethod
    def tracks_gen(self):
        """Returns a generator of tracks for each time step.

        Yields
        ------
        : :class:`datetime.datetime`
            Datetime of current time step
        : set of :class:`~.Track`
            Tracks existing in the time step
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
from abc import abstractmethod

from ..base import Base, Property
from ..models.measurement import MeasurementModel


class Sensor(Base):
    """Sensor base class

    A sensor object that operates according to a given
    :class:`~.MeasurementModel`.
    """

    measurement_model = Property(
        MeasurementModel, default=None, doc="Measurement model")

    @abstractmethod
    def gen_measurement(**kwargs):
        """Generate a measurement"""
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""Stone Soup framework: development and assessment of tracking algorithms."""
__version__ = '0.1b1'
__copyright__ = '''\
Crown Copyright (c) 2017-2019 Defence Science and Technology Laboratory UK
Crown Copyright (c) 2018-2019 Defence Research and Development Canada / Recherche et développement pour la défense Canada
Copyright (c) 2018-2019 University of Liverpool UK
'''  # noqa:E501
__license__ = 'MIT'



# -*- coding: utf-8 -*-
from functools import lru_cache

from .base import Updater
from ..base import Property
from ..resampler import Resampler
from ..types.numeric import Probability
from ..types.particle import Particle
from ..types.prediction import ParticleMeasurementPrediction
from ..types.update import ParticleStateUpdate


class ParticleUpdater(Updater):
    """Simple Particle Updater

        Perform measurement update step in the standard Kalman Filter.
        """

    resampler = Property(Resampler,
                         doc='Resampler to prevent particle degeneracy')

    def update(self, hypothesis, **kwargs):
        """Particle Filter update step

        Parameters
        ----------
        hypothesis : :class:`~.Hypothesis`
            Hypothesis with predicted state and associated detection used for
            updating.

        Returns
        -------
        : :class:`~.ParticleState`
            The state posterior
        """
        if hypothesis.measurement.measurement_model is None:
            measurement_model = self.measurement_model
        else:
            measurement_model = hypothesis.measurement.measurement_model

        for particle in hypothesis.prediction.particles:
            particle.weight *= measurement_model.pdf(
                hypothesis.measurement.state_vector, particle.state_vector,
                **kwargs)

        # Normalise the weights
        sum_w = Probability.sum(
            i.weight for i in hypothesis.prediction.particles)
        for particle in hypothesis.prediction.particles:
            particle.weight /= sum_w

        # Resample
        new_particles = self.resampler.resample(
            hypothesis.prediction.particles)

        return ParticleStateUpdate(new_particles,
                                   hypothesis,
                                   timestamp=hypothesis.measurement.timestamp)

    @lru_cache()
    def get_measurement_prediction(self, state_prediction,
                                   measurement_model=None, **kwargs):

        if measurement_model is None:
            measurement_model = self.measurement_model

        new_particles = []
        for particle in state_prediction.particles:
            new_state_vector = measurement_model.function(
                particle.state_vector, noise=0, **kwargs)
            new_particles.append(
                Particle(new_state_vector,
                         weight=particle.weight,
                         parent=particle.parent))

        return ParticleMeasurementPrediction(
            new_particles, timestamp=state_prediction.timestamp)



# -*- coding: utf-8 -*-
from ..base import Property
from .base import Type
from .hypothesis import Hypothesis
from .state import State, GaussianState, ParticleState


class Update(Type):
    """ Update type

    The base update class. Updates are returned by :class:'~.Updater' objects
    and contain the information that was used to perform the updating"""

    hypothesis = Property(Hypothesis,
                          doc="Hypothesis used for updating")


class StateUpdate(Update, State):
    """ StateUpdate type

    Most simple state update type, where everything only has time
    and a state vector. Requires a prior state that was updated,
    and the hypothesis used to update the prior.
    """


class GaussianStateUpdate(Update, GaussianState):
    """ GaussianStateUpdate type

    This is a simple Gaussian state update object, which, as the name
    suggests, is described by a Gaussian distribution.
    """


class ParticleStateUpdate(Update, ParticleState):
    """ParticleStateUpdate type

    This is a simple Particle state update object.
    """



# coding: utf-8
import pytest
import numpy as np
from scipy.stats import multivariate_normal

from ..linear import LinearGaussian


@pytest.mark.parametrize(
    "H, R, ndim_state, mapping",
    [
        (       # 1D meas, 2D state
                np.array([[1, 0]]),
                np.array([[0.1]]),
                2,
                [0],
        ),
        (       # 2D meas, 4D state
                np.array([[1, 0, 0, 0], [0, 0, 1, 0]]),
                np.diag([0.1, 0.1]),
                4,
                [0, 2],
        ),
        (       # 4D meas, 2D state
                np.array([[1, 0], [0, 0], [0, 1], [0, 0]]),
                np.diag([0.1, 0.1, 0.1, 0.1]),
                2,
                [0, None, 1, None],
        ),
    ],
    ids=["1D_meas:2D_state", "2D_meas:4D_state", "4D_meas:2D_state"]
)
def test_lgmodel(H, R, ndim_state, mapping):
    """ LinearGaussian 1D Measurement Model test """

    # State related variables
    state_vec = np.array([[n] for n in range(ndim_state)])

    # Create and a Constant Velocity model object
    lg = LinearGaussian(ndim_state=ndim_state,
                        noise_covar=R,
                        mapping=mapping)

    # Ensure ```lg.transfer_function()``` returns H
    assert np.array_equal(H, lg.matrix())

    # Ensure ```lg.covar()``` returns R
    assert np.array_equal(R, lg.covar())

    # Project a state throught the model
    # (without noise)
    meas_pred_wo_noise = lg.function(state_vec, noise=0)
    assert np.array_equal(meas_pred_wo_noise, H@state_vec)

    # Evaluate the likelihood of the predicted measurement, given the state
    # (without noise)
    prob = lg.pdf(meas_pred_wo_noise, state_vec)
    assert np.array_equal(prob, multivariate_normal.pdf(
        meas_pred_wo_noise.T,
        mean=np.array(H@state_vec).ravel(),
        cov=R).T)

    # Propagate a state vector throught the model
    # (with internal noise)
    meas_pred_w_inoise = lg.function(state_vec)
    assert not np.array_equal(meas_pred_w_inoise, H@state_vec)

    # Evaluate the likelihood of the predicted state, given the prior
    # (with noise)
    prob = lg.pdf(meas_pred_w_inoise, state_vec)
    assert np.array_equal(prob, multivariate_normal.pdf(
        meas_pred_w_inoise.T,
        mean=np.array(H@state_vec).ravel(),
        cov=R).T)

    # Propagate a state vector throught the model
    # (with external noise)
    noise = lg.rvs()
    meas_pred_w_enoise = lg.function(state_vec,
                                     noise=noise)
    assert np.array_equal(meas_pred_w_enoise, H@state_vec+noise)

    # Evaluate the likelihood of the predicted state, given the prior
    # (with noise)
    prob = lg.pdf(meas_pred_w_enoise, state_vec)
    assert np.array_equal(prob, multivariate_normal.pdf(
        meas_pred_w_enoise.T,
        mean=np.array(H@state_vec).ravel(),
        cov=R).T)



# -*- coding: utf-8 -*-
"""Test for smoother.lineargaussian"""
import datetime

import numpy as np

from ...types.detection import Detection
from ...types.hypothesis import SingleHypothesis
from ...types.state import GaussianState
from ...types.track import Track
from ...models.transition.linear import ConstantVelocity
from ...models.measurement.linear import LinearGaussian
from ...predictor.kalman import KalmanPredictor
from ...updater.kalman import KalmanUpdater
from ...smoother.lineargaussian import Backward


def test_backwards_smoother():
    """Linear Gaussian Backward Smoother test"""

    # Setup list of Detections
    T = 5
    measurements = [
        np.array([[2.486559674128609]]),
        np.array([[2.424165626519697]]),
        np.array([[6.603176662762473]]),
        np.array([[9.329099124074590]]),
        np.array([[14.637975326666801]]),
    ]
    detections = [Detection(m) for m in measurements]

    t_0 = datetime.datetime.now()
    t_delta = datetime.timedelta(0, 1)
    t = [t_0]
    for n in range(T-1):
        t.append(t[-1] + t_delta)
    for n in range(T):
        detections[n].timestamp = t[n]

    # Setup models.
    initial_state = GaussianState(np.ones([2, 1]), np.eye(2), timestamp=t_0)

    trans_model = ConstantVelocity(noise_diff_coeff=1)
    meas_model = LinearGaussian(ndim_state=2, mapping=[0],
                                noise_covar=np.array([[0.4]]))

    estimates = [initial_state]

    # Filter Initial Detection
    track = Track()
    track.append(initial_state)
    predictor = KalmanPredictor(transition_model=trans_model)
    updater = KalmanUpdater(measurement_model=meas_model)
    hypothesis = SingleHypothesis(track.state, detections[0])
    new_state = updater.update(hypothesis)
    track[0] = new_state

    # Filter Remaining Detections.
    for t in range(1, T):
        time = detections[t].timestamp
        state_pred = predictor.predict(track.state, timestamp=time)
        estimates.append(state_pred)
        hypothesis = SingleHypothesis(state_pred, detections[t])
        state_post = updater.update(hypothesis)
        track.append(state_post)

    # Smooth Track
    smoother = Backward(transition_model=trans_model)
    smoothed_track = smoother.track_smooth(track, estimates)

    smoothed_state_vectors = [
        state.state_vector for state in smoothed_track]

    # Verify Values
    target_smoothed_vectors = [
        np.array([[1.688813974839928], [1.267196351952188]]),
        np.array([[3.307200214998506], [2.187167840595264]]),
        np.array([[6.130402001958210], [3.308896367021604]]),
        np.array([[9.821303658438408], [4.119557021638030]]),
        np.array([[14.257730973981149], [4.594862462495096]])
    ]

    assert np.allclose(smoothed_state_vectors, target_smoothed_vectors)



# -*- coding: utf-8 -*-
from operator import attrgetter

import numpy as np

from ..base import Property
from .base import TrackToTrackAssociator
from ..models.measurement import MeasurementModel
from ..types.association import AssociationSet, TimeRangeAssociation
from ..types.time import TimeRange


class EuclideanTrackToTrack(TrackToTrackAssociator):
    """Euclidean track to track associator

    Compares two sets of :class:`~.tracks`, each formed of a sequence of
    :class:`~.State` objects and returns an :class:`~.Association` object for
    each time at which a the two :class:`~.State` within the :class:`~.tracks`
    are assessed to be associated.

    Associations are triggered by track states being within a threshold
    distance for a given number of timestamps. Associations are terminated when
    either the two :class:`~.tracks` end or the two :class:`~.State` are
    separated by a distance greater than the threshold at the next time step.

    Note
    ----
    Association is not prioritised based on historic associations or distance.
    If, at a specific time step, the :class:`~.State` of one of the
    :class:`~.tracks` is assessed as close to more than one track then an
    :class:`~.Association` object will be return for all possible association
    combinations
    """

    association_threshold = Property(
        float, default=10,
        doc="Threshold distance measure which states must be within for an "
            "association to be recorded.Default is 10")
    consec_pairs_confirm = Property(
        int, default=3,
        doc="Number of consecutive time instances which track pairs are "
            "required to be within a specified threshold in order for an "
            "association to be formed. Default is 3")
    consec_misses_end = Property(
        int, default=2,
        doc="Number of consecutive time instances which track pairs are "
            "required to exceed a specified threshold in order for an "
            "association to be ended. Default is 2")
    measurement_model_track1 = Property(
        MeasurementModel,
        doc="Measurement model which specifies which elements within the "
            "track state are to be used to calculate distance over")
    measurement_model_track2 = Property(
        MeasurementModel,
        doc="Measurement model which specifies which elements within the "
            "track state are to be used to calculate distance over")

    def associate_tracks(self, tracks_set_1, tracks_set_2):
        """Associate two sets of tracks together.

        Parameters
        ----------
        tracks_set_1 : list of :class:`~.Track` objects
            Tracks to associate to track set 2
        tracks_set_2 : list of :class:`~.Track` objects
            Tracks to associate to track set 1

        Returns
        -------
        AssociationSet
            Contains a set of :class:`~.Association` objects

        """
        associations = set()
        for track2 in tracks_set_2:
            truth_timestamps = [state.timestamp for state in track2.states]
            for track1 in tracks_set_1:

                track1_states = sorted(
                    (state
                     for state in track1
                     if state.timestamp in truth_timestamps),
                    key=attrgetter('timestamp'))
                track_timestamps = [state.timestamp for state in track1_states]

                track2_states = sorted(
                    (state
                     for state in track2
                     if state.timestamp in track_timestamps),
                    key=attrgetter('timestamp'))

                if not (track1_states and track2_states):
                    continue

                # At this point we should have two lists of states from
                # track1 and 2 only at the times that they both existed

                n_succesful = 0
                n_unsuccesful = 0
                start_timestamp = None
                end_timestamp = None
                # Loop through every detection pair and form associations
                for state1, state2 in zip(track1_states, track2_states):

                    distance = np.linalg.norm(
                        self.measurement_model_track1.function(
                            state1.state_vector, noise=0)
                        - self.measurement_model_track2.function(
                            state2.state_vector, noise=0))

                    if distance <= self.association_threshold:
                        n_succesful += 1
                        n_unsuccesful = 0

                        if n_succesful == 1:
                            first_timestamp = state1.timestamp
                        if n_succesful == self.consec_pairs_confirm:
                            start_timestamp = first_timestamp
                    else:
                        n_succesful = 0
                        n_unsuccesful += 1

                        if n_unsuccesful == 1:
                            end_timestamp = state1.timestamp

                        if n_unsuccesful >= self.consec_misses_end and \
                                start_timestamp:
                            associations.add(TimeRangeAssociation(
                                (track1, track2),
                                TimeRange(start_timestamp, end_timestamp)))
                            start_timestamp = None

                # close any open associations
                if start_timestamp:
                    end_timestamp = track1_states[-1].timestamp
                    associations.add(TimeRangeAssociation(
                        (track1, track2),
                        TimeRange(start_timestamp, end_timestamp)))

        return AssociationSet(associations)


class EuclideanTrackToTruth(TrackToTrackAssociator):
    """Euclidean track to truth associator

    Compares two sets of :class:`~.tracks`, each formed of a sequence of
    :class:`~.State` objects and returns an :class:`~.Association` object for
    each time at which a the two :class:`~.State` within the :class:`~.tracks`
    are assessed to be associated. Tracks are considered to be associated with
    the Truth if the true :class:`~.State` is the closest to the track and
    within the specified distance for a specified number of time steps.

    Associations between Truth and Track if the Truth is no longer the
    'closest' to the track or the distance exceeds the specified threshold for
    a specified number of consecutive time steps.

    Associates will be ended by consec_misses_end before any new associations
    are considered even if consec_pairs_confirm < consec_misses_end

    Note
    ----
    Tracks can only be associated with one Truth (one-2-one relationship) at a
    given time step however a Truth track can be associated with multiple
    Tracks (one-2-many relationship).
    """

    association_threshold = Property(
        float,
        default=10,
        doc="Threshold distance measure which states must be within for an "
            "association to be recorded.Default is 10")
    consec_pairs_confirm = Property(
        int,
        default=3,
        doc="Number of consecutive time instances which track-truth pairs are "
            "required to be within a specified threshold in order for an "
            "association to be formed. Default is 3")
    consec_misses_end = Property(
        int,
        default=2,
        doc="Number of consecutive time instances which track-truth pairs are "
            "required to exceed a specified threshold in order for an "
            "association to be ended. Default is 2")
    measurement_model_track = Property(
        MeasurementModel,
        doc="Measurement model which specifies which elements within the "
            "track state are to be used to calculate distance over")
    measurement_model_truth = Property(
        MeasurementModel,
        doc="Measurement model which specifies which elements within the "
            "truth state are to be used to calculate distance over")

    def associate_tracks(self, tracks_set, truth_set):
        """Associate Tracks

        Method compares to sets of :class:`~.Track` objects and will determine
        associations between the two sets.

        Parameters
        ----------
        tracks_set : list of :class:`~.Track` objects
            Tracks to associate to truth
        truth_set : list of :class:`~.Track` objects
            Truth to associate to tracks

        Returns
        -------
        AssociationSet
            Contains a set of :class:`~.Association` objects
        """

        associations = set()

        for track in tracks_set:

            current_truth = None
            potential_truth = None
            n_potential_successes = 0
            n_failures = 0
            potential_start_timestep = None
            start_timestamp = None
            end_timestamp = None

            for track_state in track:

                min_dist = None
                min_truth = None

                for truth in truth_set:

                    try:
                        truth_state = truth[track_state.timestamp]
                    except IndexError:
                        continue

                    distance = np.linalg.norm(
                        self.measurement_model_track.function(
                            track_state.state_vector, noise=0)
                        - self.measurement_model_truth.function(
                            truth_state.state_vector, noise=0))
                    if min_dist and distance < min_dist:
                        min_dist = distance
                        min_truth = truth
                    elif not min_dist \
                            and distance < self.association_threshold:
                        min_dist = distance
                        min_truth = truth

                # If there is not a truth track currently
                # considered to be associated to the track
                if not current_truth:
                    # If no truth associated then there's nothing to consider
                    if min_truth is None:
                        n_potential_successes = 0
                        potential_truth = None
                        potential_start_timestep = None
                    # If the latest closest truth is not being assessed
                    # as the likely truth make it so
                    elif potential_truth is not min_truth:
                        potential_truth = min_truth
                        n_potential_successes = 1
                        potential_start_timestep = track_state.timestamp

                    # Otherwise increse the number of times
                    # this truth appears in a row
                    else:
                        n_potential_successes += 1
                    # If the threshold of continuous
                    # similar matches has been made
                    if n_potential_successes >= self.consec_pairs_confirm:
                        current_truth = min_truth
                        start_timestamp = potential_start_timestep
                        end_timestamp = track_state.timestamp
                        potential_start_timestep = None
                        potential_truth = None
                        n_potential_successes = 0

                # Otherwise if there is a track currently
                # considered as the association
                else:
                    # If the closest track this time is the same
                    # update the end time (time of last association)
                    if min_truth == current_truth:
                        n_failures = 0
                        end_timestamp = track_state.timestamp
                    # Otherwise record the failed match and how
                    # many times it's been the same different
                    # potential track in a row
                    else:
                        n_failures += 1
                        if min_truth and min_truth is potential_truth:
                            n_potential_successes += 1
                        else:
                            potential_truth = min_truth
                            potential_start_timestep = track_state.timestamp
                            n_potential_successes = 1

                    # If there have been enough failed matches
                    # in a row end the association and record
                    if n_failures >= self.consec_misses_end:
                        associations.add(TimeRangeAssociation(
                            (track, current_truth),
                            TimeRange(start_timestamp, end_timestamp)))

                        # If the current potential association
                        # is strong enough to confirm then do so
                        if n_potential_successes >= self.consec_pairs_confirm:

                            current_truth = potential_truth
                            start_timestamp = potential_start_timestep
                            end_timestamp = track_state.timestamp

                        else:
                            # Otherwise wait for a new
                            # association to be good enough
                            current_truth = None
                            start_timestamp = None
                            end_timestamp = None

            # Close any open associations when the track ends
            if current_truth:

                associations.add(TimeRangeAssociation(
                    (track, current_truth),
                    TimeRange(start_timestamp, end_timestamp)))

        return AssociationSet(associations)



# -*- coding: utf-8 -*-
import numpy as np
import pytest

from ..array import StateVector, CovarianceMatrix


def test_statevector():
    with pytest.raises(ValueError):
        StateVector(np.array([0]))

    state_vector_array = np.array([[1], [2], [3], [4]])
    state_vector = StateVector(state_vector_array)

    assert np.array_equal(state_vector, state_vector_array)


def test_covariancematrix():
    """ CovarianceMatrix Type test """

    with pytest.raises(ValueError):
        CovarianceMatrix(np.array([0]))

    covar_nparray = np.array([[2.2128, 0, 0, 0],
                              [0.0002, 2.2130, 0, 0],
                              [0.3897, -0.00004, 0.0128, 0],
                              [0, 0.3897, 0.0013, 0.0135]]) * 1e3

    covar_matrix = CovarianceMatrix(covar_nparray)
    assert(np.array_equal(covar_matrix, covar_nparray))


def test_multiplication():
    vector = np.array([[1, 1, 1, 1]]).T
    state_vector = StateVector(vector)
    array = np.array([[1., 2., 3., 4.],
                      [5., 6., 7., 8.]])
    covar = CovarianceMatrix(array)

    assert np.array_equal(covar@state_vector, array@vector)
    assert np.array_equal(covar@vector, array@vector)
    assert np.array_equal(array@state_vector, array@vector)
    assert np.array_equal(state_vector.T@covar.T, vector.T@array.T)
    assert np.array_equal(vector.T@covar.T, vector.T@array.T)
    assert np.array_equal(state_vector.T@array.T, vector.T@array.T)

    assert type(array@state_vector) == type(array)
    assert type(state_vector.T@array.T) == type(state_vector)
    assert type(covar@vector) == type(covar)
    assert type(vector.T@covar.T) == type(vector)



# -*- coding: utf-8 -*-
import datetime

import pytest
import numpy as np

from ..probability import SimplePDA, JPDA
from ...types.detection import Detection, MissedDetection
from ...types.state import GaussianState
from ...types.track import Track


@pytest.fixture(params=[SimplePDA, JPDA])
def associator(request, probability_hypothesiser, probability_updater):
    if request.param is SimplePDA:
        return request.param(probability_hypothesiser)
    elif request.param is JPDA:
        return request.param(probability_hypothesiser, probability_updater, 5)


def test_probability(associator):

    timestamp = datetime.datetime.now()
    t1 = Track([GaussianState(np.array([[0]]), np.array([[1]]), timestamp)])
    t2 = Track([GaussianState(np.array([[3]]), np.array([[1]]), timestamp)])
    d1 = Detection(np.array([[1]]))
    d2 = Detection(np.array([[5]]))

    tracks = {t1, t2}
    detections = {d1, d2}

    associations = associator.associate(tracks, detections, timestamp)

    # There should be 2 associations
    assert len(associations) == 2

    # verify association probabilities are correct
    prob_t1_d1_association = [hyp.probability for hyp in associations[t1]
                              if hyp.measurement is d1]
    prob_t1_d2_association = [hyp.probability for hyp in associations[t1]
                              if hyp.measurement is d2]
    prob_t2_d1_association = [hyp.probability for hyp in associations[t2]
                              if hyp.measurement is d1]
    prob_t2_d2_association = [hyp.probability for hyp in associations[t2]
                              if hyp.measurement is d2]

    assert prob_t1_d1_association[0] > prob_t1_d2_association[0]
    assert prob_t2_d1_association[0] < prob_t2_d2_association[0]


def test_missed_detection_probability(associator):

    timestamp = datetime.datetime.now()
    t1 = Track([GaussianState(np.array([[0]]), np.array([[1]]), timestamp)])
    t2 = Track([GaussianState(np.array([[3]]), np.array([[1]]), timestamp)])
    d1 = Detection(np.array([[20]]))

    tracks = {t1, t2}
    detections = {d1}

    associations = associator.associate(tracks, detections, timestamp)

    # Best hypothesis should be missed detection hypothesis
    max_track1_prob = max([hyp.probability for hyp in associations[t1]])
    max_track2_prob = max([hyp.probability for hyp in associations[t1]])

    track1_missed_detect_prob = max(
        [hyp.probability for hyp in associations[t1]
         if isinstance(hyp.measurement, MissedDetection)])
    track2_missed_detect_prob = max(
        [hyp.probability for hyp in associations[t1]
         if isinstance(hyp.measurement, MissedDetection)])

    assert max_track1_prob == track1_missed_detect_prob
    assert max_track2_prob == track2_missed_detect_prob


def test_no_detections_probability(associator):

    timestamp = datetime.datetime.now()
    t1 = Track([GaussianState(np.array([[0]]), np.array([[1]]), timestamp)])
    t2 = Track([GaussianState(np.array([[3]]), np.array([[1]]), timestamp)])

    tracks = {t1, t2}
    detections = {}

    associations = associator.associate(tracks, detections, timestamp)

    # All hypotheses should be missed detection hypothesis
    assert all(isinstance(hypothesis.measurement, MissedDetection)
               for multihyp in associations.values()
               for hypothesis in multihyp)


def test_no_tracks_probability(associator):

    timestamp = datetime.datetime.now()
    d1 = Detection(np.array([[2]]))
    d2 = Detection(np.array([[5]]))

    tracks = {}
    detections = {d1, d2}

    associations = associator.associate(tracks, detections, timestamp)

    # Since no Tracks went in, there should be no associations
    assert not associations



# -*- coding: utf-8 -*-
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.lines

from .base import PlotGenerator
from ..base import Property
from ..types.detection import Clutter
from ..types.metric import TimeRangePlottingMetric
from ..types.prediction import Prediction
from ..types.time import TimeRange


class TwoDPlotter(PlotGenerator):
    """:class:`~.MetricManager` for the plotting data

    Plots of :class:`~.Track`, :class:`~.Detection` and
    :class:`~.GroundTruthPath` objects in two dimensions.
    """
    track_indices = Property(
        list,
        doc="Elements of track state vector to plot as x and y")
    gtruth_indices = Property(
        list,
        doc="Elements of ground truth path state vector to plot as x and y")
    detection_indices = Property(
        list,
        doc="Elements of detection state vector to plot as x and y")

    def compute_metric(self, manager, *args, **kwargs):
        """Compute the metric using the data in the metric manager

        Parameters
        ----------
        manager : MetricManager
            Containing the data to be used to create the metric(s)

        Returns
        -------
        TimeRangePlottingMetric
            contains a matplotlib figure
        """

        metric = self.plot_tracks_truth_detections(manager.tracks,
                                                   manager.groundtruth_paths,
                                                   manager.detections)
        return metric

    def plot_tracks_truth_detections(self, tracks, groundtruth_paths,
                                     detections):
        """Plots tracks, truths and detections onto a 2d matplotlib figure

        Parameters
        ----------
        tracks: set of :class:`~.Track`
            objects to be plotted as tracks
        groundtruth_paths: set of :class:`~.GroundTruthPath`
            objects to be plotted as truths
        detections: set of :class:`~.Detection`
            objects to be plotted as detections

        Returns
        ----------
        TimeRangePlottingMetric
            Contains the produced plot
        """

        fig = plt.figure()
        axis = fig.add_subplot(1, 1, 1)

        data = np.array([detection.state_vector for detection in detections
                         if not isinstance(detection, Clutter)])
        if data.any():
            axis.plot(data[:, self.detection_indices[0]],
                      data[:, self.detection_indices[1]],
                      linestyle='', marker='o')

        data = np.array([detection.state_vector for detection in
                         detections if isinstance(detection, Clutter)])
        if data.any():
            axis.plot(data[:, self.detection_indices[0]],
                      data[:, self.detection_indices[1]],
                      linestyle='', marker='2')

        for path in groundtruth_paths:
            data = np.array([state.state_vector for state in path])
            axis.plot(data[:, self.gtruth_indices[0]],
                      data[:, self.gtruth_indices[1]],
                      linestyle=':', marker='')

        for track in tracks:
            if len([state for state in track.states if not isinstance(
                    state, Prediction)]) < 2:
                continue
                # Don't plot tracks with only one detection
                #  associated; probably clutter
            data = np.array([state.state_vector for state in track.states])
            axis.plot(data[:, self.track_indices[0]],
                      data[:, self.track_indices[1]],
                      linestyle='-', marker='.')
            if hasattr(track.state, 'particles'):
                data = np.array(
                    [particle.state_vector for state in track.states for
                     particle in state.particles])
                axis.plot(data[:, self.track_indices[0]],
                          data[:, self.track_indices[1]], linestyle='',
                          marker=".", markersize=1, alpha=0.25)

        axis.set_xlabel("$x$")
        axis.set_ylabel("$y$")
        custom_legend = [
            matplotlib.lines.Line2D([0], [0], color='0', linestyle='',
                                    marker='o'),
            matplotlib.lines.Line2D([0], [0], color='0', linestyle='',
                                    marker='2'),
            matplotlib.lines.Line2D([0], [0], color='0', linestyle=':',
                                    marker=''),
            matplotlib.lines.Line2D([0], [0], color='0', linestyle='-',
                                    marker='.'),
            matplotlib.lines.Line2D([0], [0], color='0', linestyle='',
                                    marker='.', markersize=1),
        ]
        axis.legend(custom_legend,
                    ['Detections', 'Clutter', 'Path', 'Track', 'Particles'])

        timestamps = []
        for state in tracks.union(groundtruth_paths, detections):
            if state.timestamp not in timestamps:
                timestamps.append(state.timestamp)

        return TimeRangePlottingMetric(
            title='Track plot',
            value=fig,
            time_range=TimeRange(min(timestamps), max(timestamps)),
            generator=self)



# -*- coding: utf-8 -*-
import numpy as np

from ..groundtruth import GroundTruthState, GroundTruthPath


def test_groundtruthpath():
    empty_path = GroundTruthPath()

    assert len(empty_path) == 0

    groundtruth_path = GroundTruthPath([
        GroundTruthState(np.array([[0]])) for _ in range(10)])

    assert len(groundtruth_path) == 10

    state1 = GroundTruthState(np.array([[1]]))
    groundtruth_path.append(state1)
    assert groundtruth_path[-1] is state1
    assert groundtruth_path.states[-1] is state1

    state2 = GroundTruthState(np.array([[2]]))
    groundtruth_path[0] = state2
    assert groundtruth_path[0] is state2

    groundtruth_path.remove(state1)
    assert state1 not in groundtruth_path

    del groundtruth_path[0]
    assert state2 not in groundtruth_path
